# coding: utf-8
from __future__ import annotations

# http://www.numpy.org/
import numpy as np
import numpy.typing as npt

from ..tools import vect_angle

class Plane():
    """
    Class representing a plane in space. Useful for making slices.
    """
    def __init__(self,
                 normal: npt.ArrayLike,
                 point: npt.ArrayLike):
        """
        Defines a plane in space.

        Parameters
        ----------
        normal : array-like object
            3D normal vector of the plane.
        point : array-like object
            3D vector coordinate of any point in the plane.
        """

        # Set values
        self.normal = normal
        self.point = point

    @property
    def normal(self) -> np.ndarray:
        """numpy.NDArray : 3D normal unit vector of the plane."""
        return self.__normal

    @normal.setter
    def normal(self, value: npt.ArrayLike):

        # Check that value is proper
        value = np.asarray(value)
        assert value.shape == (3,), "point must be a 3D vector"

        # Save normalvector as a unit vector
        self.__normal = value / np.linalg.norm(value)

    @property
    def point(self) -> np.ndarray:
        """numpy.NDArray : a 3D vector position on the plane."""
        return self.__point

    @point.setter
    def point(self, value: npt.ArrayLike):

        # Check that value is proper
        value = np.asarray(value)
        assert value.shape == (3,), "point must be a 3D vector"

        # Save
        self.__point = value

    def below(self,
              pos: npt.ArrayLike,
              inclusive: bool = True) -> np.ndarray:
        """
        Indicates if position(s) are below the plane.  Note that identifying
        points as above or below is dependent on the sign of the plane normal.

        Parameters
        ----------
        pos : array-like object
            Nx3 array of coordinates.
        inclusive : bool, optional
            Indicates if points in the plane are to be included.
            Default value is True.

        Returns
        -------
        numpy.NDArray
            N array of bool values: True if below the plane
        """

        # Ensure pos is a numpy array
        pos = np.asarray(pos)

        # Determine above/below based on dot product with plane normal vector
        normpoint = np.dot(self.normal, self.point)
        normpos = np.inner(self.normal, pos)

        if inclusive:
            return normpos <= normpoint
        else:
            return normpos < normpoint

    def above(self,
              pos: npt.ArrayLike,
              inclusive: bool = False) -> np.ndarray:
        """
        Indicates if position(s) are above the plane.  Note that identifying
        points as above or below is dependent on the sign of the plane normal.

        Parameters
        ----------
        pos : array-like object
            Nx3 array of coordinates.
        inclusive : bool, optional
            Indicates if points in the plane are to be included.
            Default value is False.

        Returns
        -------
        numpy.NDArray
            N array of bool values: True if above the plane
        """
        return ~self.below(pos, inclusive=not inclusive)

    def operate(self,
                rotation: npt.ArrayLike,
                translation: npt.ArrayLike) -> Plane:
        """
        Return a new plane transformed by given symmetry operation

        Parameters
        ----------
        rotation: array-like, (3, 3)
            rotation matrix in cartedian coordinates
        translation: array-like, (3, )
            translation matrix in cartedian coordinates
        
        Returns
        -------
        Plane
            A new plane transformed by the specified operations.
        """
        rotation = np.array(rotation)
        translation = np.array(translation)

        new_normal = np.dot(np.linalg.inv(rotation.T), self.normal)
        new_point = np.dot(rotation, self.point) + translation
        return Plane(new_normal, new_point)

    def __eq__(self, other: Plane) -> bool:
        """
        Compare with a given plane within default tolerance parameters
        """
        if self is other:
            return True

        return self.isclose(other)

    def isclose(self,
                other: Plane,
                rtol: float = 1e-5,
                atol: float = 1e-8) -> bool:
        """
        Check the plane and a given one represent the same.
        Note that if the normal vectors of the two planes are antiparallel, they are considered to be different.

        Parameters
        ----------
        other: Plane
            a plane to be compared with
        rtol: float
            the relative tolerance parameter used in numpy.isclose
        atol: float
            the absolute tolerance parameter used in numpy.isclose

        Returns
        -------
        bool:
            Return true if the plane and a given one represent the same within tolerance.
        """
        # check if normals are parallel
        angle = vect_angle(self.normal, other.normal, unit='radian')
        if not np.isclose(angle, 0.0, rtol=rtol, atol=atol):
            return False

        # check if the point is contained in the other plane
        projected = np.dot(other.normal, self.point - other.point)
        return np.isclose(projected, 0.0, rtol=rtol, atol=atol)



# coding: utf-8

from potentials.record import Record, load_record, recordmanager
__all__ = ['Record', 'load_record', 'recordmanager']
__all__.sort()

# Add the modular Record styles
recordmanager.import_style('crystal_prototype', '.CrystalPrototype', __name__)
recordmanager.import_style('relaxed_crystal', '.RelaxedCrystal', __name__)
recordmanager.import_style('reference_crystal', '.ReferenceCrystal', __name__)
recordmanager.import_style('free_surface', '.FreeSurface', __name__)
recordmanager.import_style('stacking_fault', '.StackingFault', __name__)
recordmanager.import_style('point_defect', '.PointDefect', __name__)
recordmanager.import_style('dislocation', '.Dislocation', __name__)




# coding: utf-8

# Standard Python libraries
from warnings import warn
from typing import Optional, Tuple, Union

# http://www.numpy.org/
import numpy as np
import numpy.typing as npt

# https://www.scipy.org/
from scipy.interpolate import griddata

# https://matplotlib.org/
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
from matplotlib import cm

# atomman imports
import atomman.unitconvert as uc
from .. import Box, System
from ..tools import axes_check

__all__ = ['interpolate_contour']

def interpolate_contour(system: System,
                        prop_name: str,
                        prop_index: Union[int, tuple, None] = None,
                        prop_magnitude: bool = False,
                        prop: Optional[npt.ArrayLike] = None,  
                        plotxaxis: str = 'x',
                        plotyaxis: str = 'y',
                        xlim: Optional[tuple] = None,
                        ylim: Optional[tuple] = None,
                        zlim: Optional[tuple] = None,
                        vlim: Optional[tuple] = None,
                        xbins: int = 200,
                        ybins: int = 200,
                        dots: bool = True,
                        colorbar: bool = True,
                        vzero: bool = True,
                        title: Union[bool, str] = True,
                        length_unit: str = 'angstrom',
                        prop_unit: Optional[str] = None,
                        cmap: str = 'jet',
                        fill_value: float = np.nan,
                        figsize: Optional[tuple] = None,
                        matplotlib_axes: Optional[plt.axes] = None,
                       ) -> Tuple[float, float, Optional[plt.figure]]:
    """
    Creates a contour plot of a system's per-atom properties by interpolating
    properties between atoms.

    Parameters
    ----------
    system : atomman.System
        The system with the per-atom property that you want to plot.
    propname : str
        The name of the per-atom property that you want to plot.
    prop : array-like object, optional
        Values for the per-atom property to plot.  If not given, values will
        be taken as the "name" property of system.
    propindex : int or tuple, optional
        Specifies which component of a multidimensional property to plot.  Not
        needed if the property is scalar.
    magnitude : bool, optional
        If True, plots the per-atom magnitude of a vector property.  Cannot be
        combined with index.  Default value is False.
    plotxaxis : str or array-like object, optional
        Indicates the Cartesian direction associated with the system's atomic
        coordinates to align with the plotting x-axis.  Values are either 3D
        unit vectors, or strings 'x', 'y', or 'z' for the Cartesian axes
        directions.  plotxaxis and plotyaxis must be orthogonal.  Default value
        is 'x' = [1, 0, 0].
    plotyaxis : str or array-like object, optional
        Indicates the Cartesian direction associated with the system's atomic
        coordinates to align with the plotting y-axis.  Values are either 3D
        unit vectors, or strings 'x', 'y', or 'z' for the Cartesian axes
        directions.  plotxaxis and plotyaxis must be orthogonal.  Default value
        is 'y' = [0, 1, 0].
    xlim : tuple, optional
        The minimum and maximum coordinates along the plotting x-axis to
        include in the fit.  Values are taken in the specified length_unit.
        If not given, then the limits are set based on min and max atomic
        coordinates along the plotting axis.
    ylim : tuple, optional
        The minimum and maximum coordinates along the plotting y-axis to
        include in the fit.  Values are taken in the specified length_unit.
        If not given, then the limits are set based on min and max atomic
        coordinates along the plotting axis.
    zlim : tuple, optional
        The minimum and maximum coordinates normal to the plotting axes
        (i.e. plotxaxis X plotyaxis) to include in the fit.  Values are taken
        in the specified length_unit.  If not given, then the limits are set
        based on min and max atomic coordinates along the axis.
    vlim : tuple, optional
        Allows for the color range to be specified.  If not give, the range
        will be auto-selected based on the vzero parameter. 
    xbins : int, optional
        Specifies the number of interpolation bins to use along the plotting 
        x-axis.  Default value is 200.
    ybins : int, optional
        Specifies the number of interpolation bins to use along the plotting
        y-axis.  Default value is 200.
    dots : bool, optional
        If True, then the positions of the atoms are shown as circles.
        Default value is True.
    vzero : bool, optional
        Specifies how to auto-select the vlim values if vlim is not directly
        given. vzero=True (default) will center the range around 0, i.e.
        vlim[0] = -vlim[1].  vzero=False will select vlim[0] and vlim[1] 
        independently based on the the property values.
    colorbar: bool, optional
        If True (default) a colorbar will be added to the plot.
    title: bool or str, optional
        If True (default), a plot title will be added matching the property
        name plus index/magnitude info.  If False, no title will be used.
        If a string is given, then that will be used instead of the property
        name.
    length_unit : str, optional
        The units of length to use for the plotting x- and y-axes.  Default
        value is 'angstrom'.
    prop_unit : str or None, optional
        The units to use for the property value being plotted.  Default value
        is None, in which no unit conversion is applied.
    cmap : str, optional
        The name of the matplotlib colormap to use.  Default value is 'jet'.
    fill_value: float, optional
        Value used to fill in for grid points failed to interpolate in the fit.
        If not given, then the default is np.nan, which may cause an error in
        plotting for too narrow xlim and ylim settings.
    figsize : float or tuple, optional
        Specifies the size of the figure to create in inches.  If a single value
        is given, it will be used for the figure's width, and the height will be
        scaled based on the xlim and ylim values.  Alternatively, both the width
        and height can be set by passing a tuple of two values, but the plot will
        not be guaranteed to be "regular" with respect to length dimensions.
    matplotlib_axes : matplotlib.Axes.axes, optional
        An existing matplotlib axes object. If given, the differential displacement
        plot will be added to the specified axes of an existing figure.  This
        allows for subplots to be constructed.  Note that figsize will be ignored
        as the figure would have to be created beforehand and no automatic
        optimum scaling of the figure's dimensions will occur.
    colorbar: bool, optional
        If True (default) a colorbar will be added to the plot.
    header
    
    Returns
    -------
    intsum : float
        The area integrated sum of the property over the plotted region.
    avsum : float
        The average property value taken across all plotting bins.
    figure : matplotlib.pyplot.figure
        The generated figure object is returned if matplotlib_axes is not given.
    """
    
    # Interpret plot axis values
    plotxaxis = __plotaxisoptions(plotxaxis)
    plotyaxis = __plotaxisoptions(plotyaxis)
    
    # Build transformation matrix, T, from plot axes.
    T = axes_check([plotxaxis, plotyaxis, np.cross(plotxaxis, plotyaxis)])
    
    # Extract positions and transform using T
    pos = uc.get_in_units(np.inner(system.atoms.pos, T), length_unit)
    
    # Set plot limits
    if xlim is None:
        xlim = (pos[:, 0].min(), pos[:, 0].max())
    if ylim is None:
        ylim = (pos[:, 1].min(), pos[:, 1].max())
    if zlim is None:
        zlim = (pos[:, 2].min(), pos[:, 2].max())
        
    # Define box for identifying only points inside
    plotbox = Box(xlo = xlim[0] - 5, xhi = xlim[1] + 5,
                  ylo = ylim[0] - 5, yhi = ylim[1] + 5,
                  zlo = zlim[0], zhi = zlim[1])
    in_bounds = plotbox.inside(pos)
    
    # Manage property values
    prop_name, prop = __get_prop_values(system, prop_name, prop_index,
                                        prop_magnitude, prop, prop_unit)

    # Extract plotting coordinates and values
    x = pos[in_bounds, 0]
    y = pos[in_bounds, 1]
    c = prop[in_bounds]

    # Generate plotting grid
    grid = __grid_interpolate(x, y, c, xbins, ybins, xlim, ylim, fill_value)
    
    # Compute intsum and avsum values
    intsum = np.sum(grid)
    avsum = intsum / (xbins-1) / (ybins-1)
    intsum = intsum * (xlim[1] - xlim[0]) / (xbins-1) * (ylim[1] - ylim[0]) / (ybins-1)
    returning = [intsum, avsum]
    
    
    # Create figure if needed
    if matplotlib_axes is None:
        if figsize is None:
            figsize = (7.7, 7)
        fig = plt.figure(figsize=figsize)
        matplotlib_axes = fig.add_subplot(111)
        returning.append(fig)
    else:
        fig = matplotlib_axes.get_figure()
    
    # Get cmap from str name
    if isinstance(cmap, str):
        cmap = cm.get_cmap(cmap)
    
    if vlim is None:
        vlim = __vlim(grid, vzero)
    
    # Generate plot
    im = matplotlib_axes.imshow(grid, extent = [xlim[0], xlim[1], ylim[1], ylim[0]], 
                                cmap = cmap, norm = plt.Normalize(vmin=vlim[0], vmax=vlim[1]))
    
    # Add pretty colorbar
    if colorbar:
        ticks = np.linspace(vlim[0], vlim[1], 11, endpoint=True)
        cbar = fig.colorbar(im, ax=matplotlib_axes, fraction=0.05, pad = 0.05, ticks=ticks)
        cbar.ax.tick_params(labelsize=15)
   
    # Make axes pretty
    matplotlib_axes.axis([xlim[0], xlim[-1], ylim[0], ylim[-1]])
    #matplotlib_axes.tick_params(labelsize=15)

    # Add title
    if isinstance(title, str):
        prop_name = title
        title = True
    if title is True:
        matplotlib_axes.set_title(prop_name, size=30)

    # Add dots
    if dots:
        __adddots(matplotlib_axes, x, y, xlim, ylim)
        
    return returning

def __plotaxisoptions(plotaxis: Union[str, npt.ArrayLike]) -> np.ndarray:
    """Converts str plotaxis options to arrays as needed"""

    # Give numeric values for str plot axis terms
    if plotaxis == 'x':
        plotaxis = [1.0, 0.0, 0.0]
    elif plotaxis == 'y':
        plotaxis = [0.0, 1.0, 0.0]
    elif plotaxis == 'z':
        plotaxis = [0.0, 0.0, 1.0]

    # Convert to numpy array
    return np.asarray(plotaxis, dtype=float)


def __get_prop_values(system: System,
                      prop_name: str,
                      prop_index: Union[int, tuple, None] = None,
                      prop_magnitude: bool = False,
                      prop: Optional[npt.ArrayLike] = None,
                      prop_unit: Optional[str] = None, 
                     ) -> Tuple[str, np.ndarray]:
    """
    Manages the prop parameters
    
    Parameters
    ----------
    system : atomman.System
        The system with the per-atom property that you want to plot.
    prop_name : str
        The name of the per-atom property that you want to plot.
    prop_index : int or tuple, optional
        Specifies which component of a multidimensional property to plot.  Not
        needed if the property is scalar.
    prop_magnitude : bool, optional
        If True, plots the per-atom magnitude of a vector property.  Cannot be
        combined with index.  Default value is False.
    prop : array-like object, optional
        Values for the per-atom property to plot.  If not given, values will
        be taken as the "name" property of system.
    prop_unit : str or None, optional
        The units to use for the property value being plotted.  Default value
        is None, in which no unit conversion is applied.
        
    Returns
    -------
    prop_name : str
        The updated property name.
    prop : np.ndarray
        The per-atom property values.
    """
    
    # Get property from system
    if prop is None:
        prop = system.atoms.view[prop_name]
    
    # Handle index
    if prop_index is not None:
        if prop_magnitude is True:
            raise ValueError('prop_index and prop_magnitude cannot be combined')
        if isinstance(prop_index, (int, np.integer)):
            prop_index = [prop_index]
        else:
            prop_index = list(prop_index)
        for i in prop_index:
            prop_name += f'[{i+1}]'
        prop = prop[tuple([Ellipsis] + prop_index)]
    
    # Handle magnitude
    elif prop_magnitude is True:
        prop = np.linalg.norm(prop, axis=1)
        prop_name += '_mag'
    
    # Check that the property is of the right shape
    if prop.shape != (system.natoms, ):
        raise ValueError('property to plot must be a per-atom scalar')
        
    # Convert to the specified units
    prop = uc.get_in_units(prop, prop_unit)
    
    return prop_name, prop

def __grid_interpolate(x: npt.ArrayLike,
                       y: npt.ArrayLike,
                       v: npt.ArrayLike,
                       xbins: int = 50,
                       ybins: int = 50,
                       xlim: Optional[tuple] = None,
                       ylim: Optional[tuple] = None,
                       fill_value: float = np.nan) -> np.ndarray:
    """
    Generates 2D grid of property values by interpolating between measured
    values.

    Parameters
    ----------
    x : array-like object
        The list of x-coordinate values.
    y : array-like object
        The list of y-coordinate values.
    v : array-like object
        The list of property values associated with each (x, y) set.
    xbins : int, optional
        The number of bins to use in interpolating between the x coordinates.
        Default value is 50.
    ybins : int, optional
        The number of bins to use in interpolating between the y coordinates.
        Default value is 50.
    xlim : tuple, optional
        The minimum and maximum coordinates along the plotting x-axis to
        include in the fit.
    ylim : tuple, optional
        The minimum and maximum coordinates along the plotting y-axis to
        include in the fit.
    fill_value: float, optional
        Value used to fill in for grid points failed to interpolate in the fit.
        If not given, then the default is np.nan.

    Returns
    -------
    grid : numpy.ndarray
        The interpolated values in a grid map array.
    """
    # Generate 1D interpolation points
    xi = np.linspace(xlim[0], xlim[1], num=xbins)
    yi = np.linspace(ylim[0], ylim[1], num=ybins)

    # Generate 2D grid points
    x0, y0 = np.meshgrid(xi, yi)

    # Interpolate values to grid points
    grid = griddata((x, y), v, (x0, y0), fill_value=fill_value)
    if np.any(np.isnan(grid)):
        warn("Given xlim and ylim are too broad to interpolate. Consider shrinking xlim and ylim or set an appropriate value in fill_value.")

    return grid

def __vlim(grid: np.ndarray,
           vzero: bool = True,
           scale: float = 1) -> Tuple[float, float]:
    """
    Identifies limits to use for the property values
    """
    # Set vmin = -vmax if vzero is True
    if vzero is True:
        vmax = abs(grid).max()
        if vmax != 0.0:
            rounder = np.floor(np.log10(vmax))
            vmax = np.around(2 * vmax / 10.**rounder) * 10.**rounder / 2.
        else:
            vmax = 1e-15
        vmin = -vmax
    
    # Set vmin, vmax if vzero is False
    else:
        vmax = grid.max()
        vmin = grid.min()

        if abs(grid).max() != 0.0:
            rounder = np.floor(np.log10(grid.max() - grid.min()))

            vmax = np.around(2 * vmax / 10.**rounder) * 10.**rounder / 2.
            vmin = np.around(2 * vmin / 10.**rounder) * 10.**rounder / 2.

            if vmax == vmin:
                if vmax > 0:
                    vmin = 0
                else:
                    vmax = 0
        
        else:
            vmax = 1e-15
            vmin = -1e-15
    
    # Scale values if needed
    vmin *= scale
    vmax *= scale
    
    return vmin, vmax
    
def __adddots(ax: plt.axes,
              x: npt.ArrayLike,
              y: npt.ArrayLike,
              xlim: list,
              ylim: list):
    """
    Overlays circles onto an active 2D plot to show actual atomic positions.
    
    Parameters
    ----------
    ax : matplotlib.pyplot.axes
        The plotting axes to add the atomic positions to.
    x : array-like object
        List of x-coordinates for the atoms/dots.
    y : array-like object
        List of y-coordinates for the atoms/dots.
    xlim : list
        The [xmin, xmax] values associated with the plot.
    ylim : list
        The [ymin, ymax] values associated with the plot.
    """    
    # Set linewidth based on system dimensions
    syswidth = max([abs(xlim[1]-xlim[0]), abs(ylim[1]-ylim[0])])
    linewidth = 60. / syswidth
    
    # Add circles at each (x,y) position
    for xi, yi in zip(x, y):
        point = mpatches.Circle((xi, yi), 0.3, ls='solid', lw=linewidth)
        ax.add_artist(point)
        point.set_facecolor('none')
        point.set_edgecolor('k')


# coding: utf-8

# Standard Python imports
from typing import Optional, Union, Tuple
from pathlib import Path

from yabadaba.record import Record

# http://www.numpy.org/
import numpy as np

# https://pandas.pydata.org/
import pandas as pd

def get_crystal_prototypes(self,
                           name: Union[str, list, None] = None,
                           id: Union[str, list, None] = None,
                           key: Union[str, list, None] = None,
                           commonname: Union[str, list, None] = None,
                           prototype: Union[str, list, None] = None,
                           pearson: Union[str, list, None] = None,
                           strukturbericht: Union[str, list, None] = None,
                           sg_number: Union[int, list, None] = None,
                           sg_hm: Union[str, list, None] = None,
                           sg_schoenflies: Union[str, list, None] = None,
                           crystalfamily: Union[str, list, None] = None,
                           natypes: Union[int, list, None] = None,
                           local: Optional[bool] = None,
                           remote: Optional[bool] = None,
                           refresh_cache: bool = False,
                           return_df: bool = False,
                           verbose: bool = False
                           ) -> Union[np.ndarray, Tuple[np.ndarray, pd.DataFrame]]:
    """
    Gets all matching crystal prototypes from the database.
    
    Parameters
    ----------
    name : str or list
        The record name(s) to parse by.  For crystal prototype records, the
        names should correspond to the id.
    id : str or list, optional
        Prototype ID(s) to search for.  These are unique identifiers for each
        prototype based on comm.
    key : str or list, optional
        UUID4 key(s) to search for.  Each entry has a unique random-generated
        UUID4 key.
    commonname : str or list, optional
        Common name(s) to limit the search by.
    prototype : str or list, optional
        Prototype identifying composition(s) to limit the search by.
    pearson : str or list, optional
        The Pearson symbol(s) to limit the search by.
    strukturbericht : str or list, optional
        The strukturbericht identifier(s) to limit the search by.
    sg_number : int or list, optional
        The space group number(s) to limit the search by.
    sg_hm : str or list, optional
        The space group Hermann-Maguin identifier(s) to limit the search by.
    sg_schoenflies : str or list, optional
        The space group Schoenflies identifier(s) to limit the search by.
    crystalfamily : str, optional
        The crystal system family to limit the search by. 
    natypes : int, optional
        The number(s) of unique atom types to limit the search by.
    local : bool, optional
        Indicates if the local location is to be searched.  Default value
        matches the value set when the database was initialized.
    remote : bool, optional
        Indicates if the remote location is to be searched.  Default value
        matches the value set when the database was initialized.
    refresh_cache : bool, optional
        If the local database is of style "local", indicates if the metadata
        cache file is to be refreshed.  If False,
        metadata for new records will be added but the old record metadata
        fields will not be updated.  If True, then the metadata for all
        records will be regenerated, which is needed to update the metadata
        for modified records.
    return_df : bool, optional
        If True, then the corresponding pandas.Dataframe of metadata
        will also be returned.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    
    Returns
    -------
    numpy.ndarray
        The matching record objects
    pandas.DataFrame
        A table of the associated record metadata, returned if return_df = True.
    """

    return self.get_records(
        style='crystal_prototype', name=name, local=local, remote=remote,
        refresh_cache=refresh_cache, return_df=return_df, verbose=verbose,
        id=id, key=key, commonname=commonname, prototype=prototype,
        pearson=pearson, strukturbericht=strukturbericht,
        sg_number=sg_number, sg_hm=sg_hm, sg_schoenflies=sg_schoenflies,
        crystalfamily=crystalfamily, natypes=natypes)    

def get_crystal_prototype(self, 
                          name: Union[str, list, None] = None,
                          id: Union[str, list, None] = None,
                          key: Union[str, list, None] = None,
                          commonname: Union[str, list, None] = None,
                          prototype: Union[str, list, None] = None,
                          pearson: Union[str, list, None] = None,
                          strukturbericht: Union[str, list, None] = None,
                          sg_number: Union[int, list, None] = None,
                          sg_hm: Union[str, list, None] = None,
                          sg_schoenflies: Union[str, list, None] = None,
                          crystalfamily: Union[str, list, None] = None,
                          natypes: Union[int, list, None] = None,
                          local: Optional[bool] = None,
                          remote: Optional[bool] = None,
                          prompt: bool = True,
                          refresh_cache: bool = False,
                          verbose: bool = False) -> Record:
    
    """
    Gets exactly one matching crystal prototype from the database.
    
    Parameters
    ----------
    name : str or list
        The record name(s) to parse by.  For crystal prototype records, the
        names should correspond to the id.
    id : str or list, optional
        Prototype ID(s) to search for.  These are unique identifiers for each
        prototype based on comm.
    key : str or list, optional
        UUID4 key(s) to search for.  Each entry has a unique random-generated
        UUID4 key.
    name : str or list, optional
        Common name(s) to limit the search by.
    prototype : str or list, optional
        Prototype identifying composition(s) to limit the search by.
    pearson : str or list, optional
        The Pearson symbol(s) to limit the search by.
    strukturbericht : str or list, optional
        The strukturbericht identifier(s) to limit the search by.
    sg_number : int or list, optional
        The space group number(s) to limit the search by.
    sg_hm : str or list, optional
        The space group Hermann-Maguin identifier(s) to limit the search by.
    sg_schoenflies : str or list, optional
        The space group Schoenflies identifier(s) to limit the search by.
    crystalfamily : str or list, optional
        The crystal system family to limit the search by. 
    natypes : int or list, optional
        The number(s) of unique atom types to limit the search by.
    local : bool, optional
        Indicates if the local location is to be searched.  Default value
        matches the value set when the database was initialized.
    remote : bool, optional
        Indicates if the remote location is to be searched.  Default value
        matches the value set when the database was initialized.
    prompt : bool
        If prompt=True (default) then a screen input will ask for a selection
        if multiple matching potentials are found.  If prompt=False, then an
        error will be thrown if multiple matches are found.
    refresh_cache : bool, optional
        If the local database is of style "local", indicates if the metadata
        cache file is to be refreshed.  If False,
        metadata for new records will be added but the old record metadata
        fields will not be updated.  If True, then the metadata for all
        records will be regenerated, which is needed to update the metadata
        for modified records.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    
    Returns
    -------
    yabadaba.record.Record
        The matching record object
    """
    return self.get_record(
        style='crystal_prototype', name=name, local=local, remote=remote,
        prompt=prompt, refresh_cache=refresh_cache, verbose=verbose,
        id=id, key=key, commonname=commonname, prototype=prototype,
        pearson=pearson, strukturbericht=strukturbericht,
        sg_number=sg_number, sg_hm=sg_hm, sg_schoenflies=sg_schoenflies,
        crystalfamily=crystalfamily, natypes=natypes)

def retrieve_crystal_prototype(self,
                               name: Union[str, list, None] = None,
                               dest: Optional[Path] = None,
                               id: Union[str, list, None] = None,
                               key: Union[str, list, None] = None,
                               commonname: Union[str, list, None] = None,
                               prototype: Union[str, list, None] = None,
                               pearson: Union[str, list, None] = None,
                               strukturbericht: Union[str, list, None] = None,
                               sg_number: Union[int, list, None] = None,
                               sg_hm: Union[str, list, None] = None,
                               sg_schoenflies: Union[str, list, None] = None,
                               crystalfamily: Union[str, list, None] = None,
                               natypes: Union[int, list, None] = None,
                               local: Optional[bool] = None,
                               remote: Optional[bool] = None,
                               prompt: bool = True,
                               format: str = 'json',
                               indent: int = 4,
                               refresh_cache: bool = False,
                               verbose: bool = False):
    """
    Gets a single matching crystal prototype from the database and saves it to a
    file based on the record's name.

    Parameters
    ----------
    name : str or list, optional
        The name(s) of records to limit the search by.
    dest : path, optional
        The parent directory where the record will be saved to.  If not given,
        will use the current working directory.
    id : str or list, optional
        Prototype ID(s) to search for.  These are unique identifiers for each
        prototype based on comm.
    key : str or list, optional
        UUID4 key(s) to search for.  Each entry has a unique random-generated
        UUID4 key.
    name : str or list, optional
        Common name(s) to limit the search by.
    prototype : str or list, optional
        Prototype identifying composition(s) to limit the search by.
    pearson : str or list, optional
        The Pearson symbol(s) to limit the search by.
    strukturbericht : str or list, optional
        The strukturbericht identifier(s) to limit the search by.
    sg_number : int or list, optional
        The space group number(s) to limit the search by.
    sg_hm : str or list, optional
        The space group Hermann-Maguin identifier(s) to limit the search by.
    sg_schoenflies : str or list, optional
        The space group Schoenflies identifier(s) to limit the search by.
    crystalfamily : str or list, optional
        The crystal system family to limit the search by. 
    natypes : int or list, optional
        The number(s) of unique atom types to limit the search by.
    local : bool, optional
        Indicates if the local location is to be searched.  Default value
        matches the value set when the database was initialized.
    remote : bool, optional
        Indicates if the remote location is to be searched.  Default value
        matches the value set when the database was initialized.
    prompt : bool, optional
        If prompt=True (default) then a screen input will ask for a selection
        if multiple matching potentials are found.  If prompt=False, then an
        error will be thrown if multiple matches are found.
    format : str, optional
        The file format to save the record in: 'json' or 'xml'.  Default
        is 'json'.
    indent : int, optional
        The number of space indentation spacings to use in the saved
        record for the different tiered levels.  Default is 4.  Giving None
        will create a compact record.
    refresh_cache : bool, optional
        If the local database is of style "local", indicates if the metadata
        cache file is to be refreshed.  If False,
        metadata for new records will be added but the old record metadata
        fields will not be updated.  If True, then the metadata for all
        records will be regenerated, which is needed to update the metadata
        for modified records.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    
    Raises
    ------
    ValueError
        If local or remote is set to True when the corresponding database
        interaction has not been set.
    ValueError
        If multiple or no matching records are discovered.
    """
    self.retrieve_record(
        style='crystal_prototype', name=name, dest=dest, local=local,
        remote=remote, prompt=prompt, format=format,
        indent=indent, refresh_cache=refresh_cache, verbose=verbose,
        id=id, key=key, commonname=commonname, prototype=prototype,
        pearson=pearson, strukturbericht=strukturbericht,
        sg_number=sg_number, sg_hm=sg_hm, sg_schoenflies=sg_schoenflies,
        crystalfamily=crystalfamily, natypes=natypes)

def download_crystal_prototypes(self,
                                name: Union[str, list, None] = None,
                                id: Union[str, list, None] = None,
                                key: Union[str, list, None] = None,
                                commonname: Union[str, list, None] = None,
                                prototype: Union[str, list, None] = None,
                                pearson: Union[str, list, None] = None,
                                strukturbericht: Union[str, list, None] = None,
                                sg_number: Union[int, list, None] = None,
                                sg_hm: Union[str, list, None] = None,
                                sg_schoenflies: Union[str, list, None] = None,
                                crystalfamily: Union[str, list, None] = None,
                                natypes: Union[int, list, None] = None,
                                overwrite: bool = False,
                                return_records: bool = False,
                                verbose: bool = False) -> Optional[np.ndarray]:
    """
    Downloads crystal prototypes from the remote to the local.
    
    Parameters
    ----------
    name : str or list
        The record name(s) to parse by.  For crystal prototype records, the
        names should correspond to the id.
    id : str or list, optional
        Prototype ID(s) to search for.  These are unique identifiers for each
        prototype based on comm.
    key : str or list, optional
        UUID4 key(s) to search for.  Each entry has a unique random-generated
        UUID4 key.
    name : str or list, optional
        Common name(s) to limit the search by.
    prototype : str or list, optional
        Prototype identifying composition(s) to limit the search by.
    pearson : str or list, optional
        The Pearson symbol(s) to limit the search by.
    strukturbericht : str or list, optional
        The strukturbericht identifier(s) to limit the search by.
    sg_number : int or list, optional
        The space group number(s) to limit the search by.
    sg_hm : str or list, optional
        The space group Hermann-Maguin identifier(s) to limit the search by.
    sg_schoenflies : str or list, optional
        The space group Schoenflies identifier(s) to limit the search by.
    crystalfamily : str or list, optional
        The crystal system families to limit the search by. 
    natypes : int or list, optional
        The number(s) of unique atom types to limit the search by.
    overwrite : bool, optional
        Flag indicating if any existing local records with names matching
        remote records are updated (True) or left unchanged (False).  Default
        value is False.
    return_records : bool, optional
        If True, the retrieved record objects are also returned.  Default
        value is False.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.

    Returns
    -------
    numpy.ndarray
        The matching records - returned if return_records=True
    """

    return self.download_records(
        style='crystal_prototype', name=name, overwrite=overwrite,
        return_records=return_records, verbose=verbose,
        id=id, key=key, commonname=commonname, prototype=prototype,
        pearson=pearson, strukturbericht=strukturbericht,
        sg_number=sg_number, sg_hm=sg_hm, sg_schoenflies=sg_schoenflies,
        crystalfamily=crystalfamily, natypes=natypes)


# coding: utf-8

# Standard Python libraries
import io
from typing import Optional, Union

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

# https://github.com/usnistgov/potentials
from potentials import load_record
from potentials.record import BasePotentalLAMMPS

def Potential(model: Union[str, io.IOBase],
              name: Optional[str] = None,
              pot_dir: Optional[str] = None,
              kim_id: Optional[str] = None,
              potkey: Optional[str] = None,
              potid: Optional[str] = None,
              symbolset: Union[str, list, None] = None
              ) -> BasePotentalLAMMPS:
    """
    Class initializer switch for PotentialLAMMPS or PotentialLAMMPSKIM objects.

    Parameters
    ----------
    model : str or file-like object
        A JSON/XML data model containing a potential-LAMMPS or a
        potential-LAMMPS-KIM branch.
    name : str, optional
        The record name to use.  If not given, this will be set to the
        potential's id.
    pot_dir : str, optional
        The path to a directory containing any artifacts associated with
        the potential.  Default value is None, which assumes any required
        files will be in the working directory when LAMMPS is executed.
    kim_id : str, optional
        The full KIM model id indicating the version of a KIM model to use.
        If not given, then the newest known version will be used.
    potkey : str, optional
        Specifies which potential (by potkey value) to use.  Only important
        if it is a kim model associated with multiple potential entries.
        This controls which symbols are available.
    potid : str, optional
        Specifies which potential (by potid value) to use.  Only important
        if it is a kim model associated with multiple potential entries.
        This controls which symbols are available.
    symbolset : str or list, optional
        Specifies which potential (by symbols value) to use.  Only important
        if it is a kim model associated with multiple potential entries.
        If potkey or potid is not given, then the first potential entry
        found with all listed symbols will be selected.
    
    Returns
    -------
    potentials.record.BasePotentalLAMMPS
        The matching LAMMPS Potential object
    """
    model = DM(model)
    try:
        # Search for potential-LAMMPS branch
        model.find('potential-LAMMPS')
    except ValueError:
        try:
            # Search for potential-LAMMPS-KIM branch
            model.find('potential-LAMMPS-KIM')
        except ValueError as err:
            raise ValueError('Failed to find either potential-LAMMPS or potential-LAMMPS-KIM content') from err
        else:
            return load_record('potential_LAMMPS_KIM', model=model, name=name, id=kim_id,
                               potkey=potkey, potid=potid, symbolset=symbolset)
    else:
        return load_record('potential_LAMMPS', model=model, name=name, pot_dir=pot_dir)


# coding: utf-8
 
# Standard Python imports
from __future__ import annotations
from typing import Optional, Union

# atomman imports
from ... import System
from ...library import Database

def load(name: Union[str, list, None] = None,
         id: Union[str, list, None] = None,
         key: Union[str, list, None] = None,
         commonname: Union[str, list, None] = None,
         prototype: Union[str, list, None] = None,
         pearson: Union[str, list, None] = None,
         strukturbericht: Union[str, list, None] = None,
         sg_number: Union[int, list, None] = None,
         sg_hm: Union[str, list, None] = None,
         sg_schoenflies: Union[str, list, None] = None,
         a: Optional[float] = None,
         b: Optional[float] = None,
         c: Optional[float] = None,
         alpha: Optional[float] = None,
         beta: Optional[float] = None,
         gamma: Optional[float] = None,
         symbols: Optional[tuple] = None,
         database: Optional[Database] = None,
         local: Optional[bool] = True,
         remote: Optional[bool] = True,
         prompt: bool = True,
         refresh_cache: bool = False,
         verbose: bool = False) -> System:
    """
    Loads a crystal prototype record from the library. If multiple matches
    are found based on inputs a selection menu will appear.
    
    name : str or list, optional.
        Record names to search for.  These should be the same values as id.
    id : str or list, optional
        Prototype ID(s) to search for.  These are unique identifiers for each
        prototype based on comm.
    key : str or list, optional
        UUID4 key(s) to search for.  Each entry has a unique random-generated
        UUID4 key.
    commonname : str or list, optional
        Common name(s) to limit the search by.
    prototype : str or list, optional
        Prototype identifying composition(s) to limit the search by.
    pearson : str or list, optional
        The Pearson symbol(s) to limit the search by.
    strukturbericht : str or list, optional
        The strukturbericht identifier(s) to limit the search by.
    sg_number : int or list, optional
        The space group number(s) to limit the search by.
    sg_hm : str or list, optional
        The space group Hermann-Maguin identifier(s) to limit the search by.
    sg_schoenflies : str or list, optional
        The space group Schoenflies identifier(s) to limit the search by.
    a : float, optional
        The a lattice parameter to scale the prototype by. Can only be given
        if it is a unique lattice parameter for the prototype's crystal family,
        and if all other unique lattice parameters are given.
    b : float, optional
        The b lattice parameter to scale the prototype by. Can only be given
        if it is a unique lattice parameter for the prototype's crystal family,
        and if all other unique lattice parameters are given.
    c : float, optional
        The c lattice parameter to scale the prototype by. Can only be given
        if it is a unique lattice parameter for the prototype's crystal family,
        and if all other unique lattice parameters are given.
    alpha : float, optional
        The alpha lattice angle to scale the prototype by. Can only be given
        if it is a unique lattice parameter for the prototype's crystal family,
        and if all other unique lattice parameters are given.
    beta : float, optional
        The beta lattice angle to scale the prototype by. Can only be given
        if it is a unique lattice parameter for the prototype's crystal family,
        and if all other unique lattice parameters are given.
    gamma : gamma, optional
        The alpha lattice angle to scale the prototype by. Can only be given
        if it is a unique lattice parameter for the prototype's crystal family,
        and if all other unique lattice parameters are given.
    symbols : tuple, optional
        Allows the list of element symbols to be assigned during loading.
    database : atomman.library.Database, optional
        A pre-defined Database object to use.  If not given, will initialize
        a new Database object.  Passing in a database can save time if multiple
        calls are made for the same record type. 
    local : bool, optional
        Indicates if the Database object is to look for local records.  Default
        is True.  Ignored if database is given.
    remote : bool, optional
        Indicates if the Database object is to look for remote records.  Default
        is True.  Ignored if database is given.
    prompt : bool, optional
        If prompt=True (default) then a screen input will ask for a selection
        if multiple matching potentials are found.  If prompt=False, then an
        error will be thrown if multiple matches are found.
    refresh_cache : bool, optional
        If the local database is of style "local", indicates if the metadata
        cache file is to be refreshed.  If False,
        metadata for new records will be added but the old record metadata
        fields will not be updated.  If True, then the metadata for all
        records will be regenerated, which is needed to update the metadata
        for modified records.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    
    Returns
    -------
    system : atomman.System
        The system object generated from the crystal prototype.
    """

    # Create Database object and load if needed
    if database is None:
        database = Database.Database()
    
    # Get crystal prototype record
    prototype = database.get_crystal_prototype(local=local, remote=remote, name=name,
                            id=id, key=key, commonname=commonname, prototype=prototype,
                            pearson=pearson, strukturbericht=strukturbericht,
                            sg_number=sg_number, sg_hm=sg_hm, sg_schoenflies=sg_schoenflies,
                            prompt=prompt, refresh_cache=refresh_cache, verbose=verbose)
    
    # Retrieve unit cell information and set symbols
    ucell = prototype.ucell
    ucell.symbols = symbols
    
    # Allow lattice constants to be set based on crystal family
    family = ucell.box.identifyfamily()
    if family == 'cubic':
        if a is not None:
            if b is not None or c is not None or alpha is not None or beta is not None or gamma is not None:
                raise ValueError('Only a can be set for cubic prototypes')
            ucell.box_set(a=a, b=a, c=a, scale=True)
            
    elif family == 'hexagonal':
        if a is not None and c is not None:
            if b is not None or alpha is not None or beta is not None or gamma is not None:
                raise ValueError('Only a, c can be set for hexagonal prototypes')
            ucell.box_set(a=a, b=a, c=c, gamma=120, scale=True)
        elif a is not None or c is not None:
            raise ValueError('All or neither of a, c must be set for hexagonal prototypes')
            
    elif family == 'tetragonal':
        if a is not None and c is not None:
            if b is not None or alpha is not None or beta is not None or gamma is not None:
                raise ValueError('Only a, c can be set for tetragonal prototypes')
            ucell.box_set(a=a, b=a, c=c, scale=True)
        elif a is not None or c is not None:
            raise ValueError('All or neither of a, c must be set for tetragonal prototypes')
            
    elif family == 'rhombohedral':
        if a is not None and alpha is not None:
            if b is not None or c is not None or beta is not None or gamma is not None:
                raise ValueError('Only a, alpha can be set for rhombohedral prototypes')
            ucell.box_set(a=a, b=a, c=a, alpha=alpha, beta=alpha, gamma=alpha, scale=True)
        elif a is not None or alpha is not None:
            raise ValueError('All or neither of a, alpha must be set for rhombohedral prototypes')
            
    elif family == 'orthorhombic':
        if a is not None and b is not None and c is not None:
            if alpha is not None or beta is not None or gamma is not None:
                raise ValueError('Only a, b, c can be set for orthorhombic prototypes')
            ucell.box_set(a=a, b=b, c=c, scale=True)
        elif a is not None or b is not None or c is not None:
            raise ValueError('All or neither of a, b, c must be set for orthorhombic prototypes')
            
    elif family == 'monoclinic':
        if a is not None and b is not None and c is not None and beta is not None:
            if alpha is not None or gamma is not None:
                raise ValueError('Only a, b, c, beta can be set for monoclinic prototypes')
            ucell.box_set(a=a, b=b, c=c, beta=beta, scale=True)
        elif a is not None or b is not None or c is not None or beta is not None:
            raise ValueError('All or neither of a, b, c, beta must be set for monoclinic prototypes')
            
    else:
        if a is not None and b is not None and c is not None and alpha is not None and beta is not None and gamma is not None:
            ucell.box_set(a=a, b=b, c=c, alpha=alpha, beta=beta, gamma=gamma, scale=True)
        elif a is not None or b is not None or c is not None or alpha is not None or beta is not None or gamma is not None:
            raise ValueError('All or neither of a, b, c, alpha, beta, gamma must be set for triclinic prototypes')
    
    return ucell


# coding: utf-8

# Standard Python imports
from typing import Callable, Union

# http://www.numpy.org/
import numpy as np
import numpy.typing as npt

def euler(ratefxn: Callable,
          coord: npt.ArrayLike,
          timestep: float,
          **kwargs) -> Union[float, np.ndarray]:
    """
    Performs Euler ODE integration for a timestep.
    
    Parameters
    ----------
    ratefxn : function
        The rate function to use.  Should be a function of coord.
    coord : array-like object
        The coordinate(s) of the last timestep.
    timestep : float
        The timestep value to use.
    **kwargs : any
        Any extra keyword parameters to pass on to ratefxn.
    
    Returns
    -------
    float or numpy.ndarray
        The coordinate(s) moved forward by timestep.
    """
    
    return coord + timestep * ratefxn(coord, **kwargs)


# coding: utf-8

# Standard Python imports
from importlib import resources

# potentials imports
from potentials import build_lammps_potential, settings

# atomman imports
from . import unitconvert
from . import tools
from . import mep
from . import region
from . import lammps
from .dump import *
from .dump import __all__ as dump_all
from .core import *
from .core import __all__ as core_all
from . import cluster
from . import library
from .library import load_lammps_potential
from .load import *
from .load import __all__ as load_all
from . import plot
from . import defect

# Read version from VERSION file
__version__ = resources.read_text('atomman', 'VERSION').strip()

# Build all list
__all__ = ['__version__'] + dump_all + core_all + load_all
__all__ += ['load_lammps_potential', 'build_lammps_potential', 'settings']
__all__.sort()

# Define default working units
unitconvert.reset_units(length = 'angstrom', mass = 'amu', energy='eV', charge='e')


# coding: utf-8

# Standard Python imports
import io
from typing import Optional, Union, Tuple

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

# https://github.com/usnistgov/yabadaba
from yabadaba.record import Record
from yabadaba import load_query

# https://pandas.pydata.org/
import pandas as pd
class FreeSurface(Record):
    """
    Class for representing free_surface records, which collect the parameters
    necessary for atomman to generate a particular free surface.
    """
    def __init__(self,
                 model: Union[str, io.IOBase, DM, None] = None,
                 name: Optional[str] = None):
        """
        Initializes a Record object for a given style.
        
        Parameters
        ----------
        model : str, file-like object, DataModelDict
            The contents of the record.
        name : str, optional
            The unique name to assign to the record.  If model is a file
            path, then the default record name is the file name without
            extension.
        """
        if model is not None:
            super().__init__(model=model, name=name)
        elif name is not None:
            self.name = name

    @property
    def style(self) -> str:
        """str: The record style"""
        return 'free_surface'

    @property
    def xsd_filename(self) -> Tuple[str, str]:
        """tuple: The module path and file name of the record's xsd schema"""
        return ('atomman.library.xsd', f'{self.style}.xsd')

    @property
    def modelroot(self) -> str:
        """str: The root element of the content"""
        return 'free-surface'
    
    @property
    def key(self) -> str:
        """str : A UUID4 key assigned to the record"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__key

    @key.setter
    def key(self, value: str):
        self.__key = str(value)
    
    @property
    def id(self) -> str:
        """str : A unique id assigned to the record"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__id

    @id.setter
    def id(self, value: str):
        self.__id = str(value)

    @property
    def family(self) -> str:
        """str : The prototype/reference id the defect is defined for"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__family

    @family.setter
    def family(self, value: str):
        self.__family = str(value)

    @property
    def parameters(self) -> dict:
        """dict : Defect parameters for atomman structure generator"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__parameters

    def load_model(self,
                   model: Union[str, io.IOBase, DM],
                   name: Optional[str] = None):
        """
        Loads record contents from a given model.

        Parameters
        ----------
        model : str or DataModelDict
            The model contents of the record to load.
        name : str, optional
            The name to assign to the record.  Often inferred from other
            attributes if not given.
        """
        super().load_model(model, name=name)
        content = self.model[self.modelroot]

        self.key = content['key']
        self.id = content['id']
        self.family = content['system-family']
        self.__parameters = dict(content['calculation-parameter'])

    def build_model(self) -> DM:
        """
        Returns the object info as data model content
        
        Returns
        ----------
        DataModelDict
            The data model content.
        """
        model = DM()
        model[self.modelroot] = content = DM()

        content['key'] = self.key
        content['id'] = self.id
        content['system-family'] = self.family
        content['calculation-parameter'] = DM(self.parameters)

        self._set_model(model)
        return model

    def metadata(self) -> dict:
        """
        Generates a dict of simple metadata values associated with the record.
        Useful for quickly comparing records and for building pandas.DataFrames
        for multiple records of the same style.
        """
        meta = {}
        meta['name'] = self.name
        meta['id'] = self.id
        meta['family'] = self.family
        meta['hkl'] = self.parameters['hkl']
        if 'shiftindex' in self.parameters:
            meta['shiftindex'] = self.parameters['shiftindex']
        meta['cutboxvector'] = self.parameters['cutboxvector']
        
        return meta

    @property
    def queries(self) -> dict:
        """dict: Query objects and their associated parameter names."""
        return {
            'key': load_query(
                style='str_match',
                name='key', 
                path=f'{self.modelroot}.key'),
            'id': load_query(
                style='str_match',
                name='id',
                path=f'{self.modelroot}.id'),
            'family': load_query(
                style='str_match',
                name='family',
                path=f'{self.modelroot}.system-family'),
            'hkl': load_query(
                style='str_match',
                name='hkl',
                path=f'{self.modelroot}.calculation-parameter.hkl'),
            'shiftindex': load_query(
                style='int_match',
                name='shiftindex',
                path=f'{self.modelroot}.calculation-parameter.shiftindex'),
            'cutboxvector': load_query(
                style='str_match',
                name='cutboxvector',
                path=f'{self.modelroot}.calculation-parameter.cutboxvector'),
        }

    def pandasfilter(self,
                     dataframe: pd.DataFrame,
                     name: Union[str, list, None] = None,
                     key: Union[str, list, None] = None,
                     id: Union[str, list, None] = None,
                     family: Union[str, list, None] = None,
                     hkl: Union[str, list, None] = None,
                     shiftindex: Union[int, list, None] = None,
                     cutboxvector: Union[str, list, None] = None) -> pd.Series:
        """
        Filters a pandas.DataFrame based on kwargs values for the record style.
        
        Parameters
        ----------
        dataframe : pandas.DataFrame
            A table of metadata for multiple records of the record style.
        name : str or list
            The record name(s) to parse by.
        id : str or list
            The record id(s) to parse by.
        key : str or list
            The record key(s) to parse by.
        family : str or list
            Parent prototype/reference id(s) to parse by.
        hkl : str or list
            Space delimited plane(s) to parse by.
        shiftindex : int or list
            The shift index indicating the termination plane.
        cutboxvector : str or list
            cutboxvector value(s) to parse by.
        
        Returns
        -------
        pandas.Series
            Boolean map of matching values
        """
        matches = super().pandasfilter(dataframe, name=name, key=key, id=id,
                                       family=family, hkl=hkl, shiftindex=shiftindex,
                                       cutboxvector=cutboxvector)
        return matches

    def mongoquery(self,
                   name: Union[str, list, None] = None,
                   key: Union[str, list, None] = None,
                   id: Union[str, list, None] = None,
                   family: Union[str, list, None] = None,
                   hkl: Union[str, list, None] = None,
                   shiftindex: Union[int, list, None] = None,
                   cutboxvector: Union[str, list, None] = None) -> dict:
        """
        Builds a Mongo-style query based on kwargs values for the record style.
        
        Parameters
        ----------
        name : str or list
            The record name(s) to parse by.
        id : str or list
            The record id(s) to parse by.
        key : str or list
            The record key(s) to parse by.
        family : str or list
            Parent prototype/reference id(s) to parse by.
        hkl : str or list
            Space delimited plane(s) to parse by.
        cutboxvector : str or list
            cutboxvector value(s) to parse by.
        
        Returns
        -------
        dict
            The Mongo-style query
        """   
        mquery = super().mongoquery(name=name, key=key, id=id,
                                    family=family, hkl=hkl, shiftindex=shiftindex,
                                    cutboxvector=cutboxvector)
        return mquery

    def cdcsquery(self,
                  key: Union[str, list, None] = None,
                  id: Union[str, list, None] = None,
                  family: Union[str, list, None] = None,
                  hkl: Union[str, list, None] = None,
                  shiftindex: Union[int, list, None] = None,
                  cutboxvector: Union[str, list, None] = None) -> dict:
        """
        Builds a CDCS-style query based on kwargs values for the record style.
        
        Parameters
        ----------
        id : str or list
            The record id(s) to parse by.
        key : str or list
            The record key(s) to parse by.
        family : str or list
            Parent prototype/reference id(s) to parse by.
        hkl : str or list
            Space delimited plane(s) to parse by.
        cutboxvector : str or list
            cutboxvector value(s) to parse by.
        
        Returns
        -------
        dict
            The CDCS-style query
        """
        mquery = super().cdcsquery(key=key, id=id,
                                    family=family, hkl=hkl, shiftindex=shiftindex,
                                    cutboxvector=cutboxvector)
        return mquery


# coding: utf-8
from .dvect import dvect
from .dmag import dmag 
from .nlist import nlist 
from .NeighborList import NeighborList
from .Atoms import Atoms
from .Box import Box
from .ElasticConstants import ElasticConstants
from .System import System
from .displacement import displacement

__all__ = ['displacement', 'dvect', 'dmag', 'nlist', 'Atoms', 'Box',
           'ElasticConstants', 'NeighborList', 'System']


# coding: utf-8

# Standard Python imports
from pathlib import Path
from typing import Optional, Union, Tuple

from yabadaba.record import Record

# http://www.numpy.org/
import numpy as np

# https://pandas.pydata.org/
import pandas as pd

# https://docs.python-requests.org/
import requests

# atomman imports
from ...load import load_pymatgen_Structure, load_poscar
from ..record import load_record
from ...tools import aslist

def get_reference_crystals(self,
                           name: Union[str, list, None] = None,
                           key: Union[str, list, None] = None,
                           id: Union[str, list, None] = None,
                           sourcename: Union[str, list, None] = None,
                           sourcelink: Union[str, list, None] = None,
                           crystalfamily: Union[str, list, None] = None,
                           composition: Union[str, list, None] = None,
                           symbols: Union[str, list, None] = None,
                           natoms: Union[int, list, None] = None,
                           natypes: Union[int, list, None] = None,
                           local: Optional[bool] = None,
                           remote: Optional[bool] = None, 
                           refresh_cache: bool = False,
                           return_df: bool = False,
                           verbose: bool = False
                           ) -> Union[np.ndarray, Tuple[np.ndarray, pd.DataFrame]]:
    """
    Gets all matching reference crystals from the database.
    
    Parameters
    ----------
    name : str or list, optional
        The record name(s) to parse by.  For reference crystal records, the
        names should correspond to the id.
    id : str or list, optional
        The record id(s) to parse by.  For reference crystal records, the
        id are letters identifying the source database "mp-", "mvc-", or "oqmd-",
        followed by the source database's identification number.
    key : str or list, optional
        UUID4 key(s) to search for.  Each entry has a unique random-generated
        UUID4 key.
    sourcename : str or list, optional
        The full name of source DFT databases to limit the search by.
        "Materials Project" or "Open Quantum Materials Database".            
    sourcelink : str or list, optional
        The web link of the source DFT databases to limit the search by.
    crystalfamily : str or list, optional
        The crystal system families to limit the search by.
    composition : str or list, optional
        The reduced compositions of the structures to limit the search by.
        Element symbols are sorted alphabetically.
    symbols : str or list, optional
        Element symbols in the crystal to limit the search by.
    natoms : int or list, optional
        The number of unique atoms in the crystal's unit cell to limit the
        search by.
    natypes : int or list, optional
        The number(s) of unique atom types to limit the search by.
    local : bool, optional
        Indicates if the local location is to be searched.  Default value
        matches the value set when the database was initialized.
    remote : bool, optional
        Indicates if the remote location is to be searched.  Default value
        matches the value set when the database was initialized.
    refresh_cache : bool, optional
        If the local database is of style "local", indicates if the metadata
        cache file is to be refreshed.  If False,
        metadata for new records will be added but the old record metadata
        fields will not be updated.  If True, then the metadata for all
        records will be regenerated, which is needed to update the metadata
        for modified records.
    return_df : bool, optional
        If True, then the corresponding pandas.Dataframe of metadata
        will also be returned.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    
    Returns
    -------
    numpy.ndarray
        The matching record objects
    pandas.DataFrame
        A table of the associated record metadata, returned if return_df = True.
    """

    return self.get_records(
        style='reference_crystal', name=name, local=local, remote=remote, 
        refresh_cache=refresh_cache, return_df=return_df, verbose=verbose,
        key=key, id=id, sourcename=sourcename, sourcelink=sourcelink,
        crystalfamily=crystalfamily, composition=composition,
        symbols=symbols, natoms=natoms, natypes=natypes)

def promptfxn(df):
    """
    promptfxn for reference crystals
    """
    header = '#  id           comp     c-family     natoms alat'
    print(header)
    
    js = df.sort_values(['composition', 'natoms', 'crystalfamily', 'a']).index
    for i, j in enumerate(js):
        crystal = df.loc[j]
        row =  f'{i+1:2} {crystal.id:12.12} '
        row += f'{crystal.composition:8.8} '
        row += f'{crystal.crystalfamily:12.12} '
        row += f'{crystal.natoms:5d} '
        row += f'{crystal.a:7.4f}'
        print(row)
    
    i = int(input('Please select one:')) - 1
    if i < 0 or i >= len(js):
        raise ValueError('Invalid selection')

    return js[i]

def get_reference_crystal(self, 
                          name: Union[str, list, None] = None,
                          key: Union[str, list, None] = None,
                          id: Union[str, list, None] = None,
                          sourcename: Union[str, list, None] = None,
                          sourcelink: Union[str, list, None] = None,
                          crystalfamily: Union[str, list, None] = None,
                          composition: Union[str, list, None] = None,
                          symbols: Union[str, list, None] = None,
                          natoms: Union[int, list, None] = None,
                          natypes: Union[int, list, None] = None,
                          local: Optional[bool] = None,
                          remote: Optional[bool] = None,
                          prompt: bool = True,
                          refresh_cache: bool = False,
                          verbose: bool = False) -> Record:
    """
    Gets exactly one matching reference crystal from the database.
    
    Parameters
    ----------
    local : bool, optional
        Indicates if the local location is to be searched.  Default value
        matches the value set when the database was initialized.
    remote : bool, optional
        Indicates if the remote location is to be searched.  Default value
        matches the value set when the database was initialized.
    name : str or list, optional
        The record name(s) to parse by.  For reference crystal records, the
        names should correspond to the id.
    id : str or list, optional
        The record id(s) to parse by.  For reference crystal records, the
        id are letters identifying the source database "mp-", "mvc-", or "oqmd-",
        followed by the source database's identification number.
    key : str or list, optional
        UUID4 key(s) to search for.  Each entry has a unique random-generated
        UUID4 key.
    sourcename : str or list, optional
        The full name of source DFT databases to limit the search by.
        "Materials Project" or "Open Quantum Materials Database".            
    sourcelink : str or list, optional
        The web link of the source DFT databases to limit the search by.
    crystalfamily : str or list, optional
        The crystal system families to limit the search by.
    composition : str or list, optional
        The reduced compositions of the structures to limit the search by.
        Element symbols are sorted alphabetically.
    symbols : str or list, optional
        Element symbols in the crystal to limit the search by.
    natoms : int or list, optional
        The number of unique atoms in the crystal's unit cell to limit the
        search by.
    natypes : int or list, optional
        The number(s) of unique atom types to limit the search by.
    prompt : bool
        If prompt=True (default) then a screen input will ask for a selection
        if multiple matching potentials are found.  If prompt=False, then an
        error will be thrown if multiple matches are found.
    refresh_cache : bool, optional
        If the local database is of style "local", indicates if the metadata
        cache file is to be refreshed.  If False,
        metadata for new records will be added but the old record metadata
        fields will not be updated.  If True, then the metadata for all
        records will be regenerated, which is needed to update the metadata
        for modified records.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.

    Returns
    -------
    yabadaba.record.Record
        The matching record object
    """
    return self.get_record(
        style='reference_crystal', name=name, local=local, remote=remote, prompt=prompt,
        promptfxn=promptfxn, refresh_cache=refresh_cache, verbose=verbose,
        key=key, id=id, sourcename=sourcename, sourcelink=sourcelink,
        crystalfamily=crystalfamily, composition=composition,
        symbols=symbols, natoms=natoms, natypes=natypes)

def retrieve_reference_crystal(self,
                               name: Union[str, list, None] = None,
                               dest: Optional[Path] = None,
                               key: Union[str, list, None] = None,
                               id: Union[str, list, None] = None,
                               sourcename: Union[str, list, None] = None,
                               sourcelink: Union[str, list, None] = None,
                               crystalfamily: Union[str, list, None] = None,
                               composition: Union[str, list, None] = None,
                               symbols: Union[str, list, None] = None,
                               natoms: Union[int, list, None] = None,
                               natypes: Union[int, list, None] = None,
                               local: Optional[bool] = None,
                               remote: Optional[bool] = None, 
                               prompt: bool = True,
                               format: str = 'json',
                               indent: int = 4,
                               refresh_cache: bool = False,
                               verbose: bool = False):
    """
    Gets a single matching relaxed crystal from the database and saves it to a
    file based on the record's name.

    Parameters
    ----------
    name : str or list, optional
        The name(s) of records to limit the search by.
    dest : path, optional
        The parent directory where the record will be saved to.  If not given,
        will use the current working directory.
    id : str or list, optional
        The record id(s) to parse by.  For reference crystal records, the
        id are letters identifying the source database "mp-", "mvc-", or "oqmd-",
        followed by the source database's identification number.
    key : str or list, optional
        UUID4 key(s) to search for.  Each entry has a unique random-generated
        UUID4 key.
    sourcename : str or list, optional
        The full name of source DFT databases to limit the search by.
        "Materials Project" or "Open Quantum Materials Database".            
    sourcelink : str or list, optional
        The web link of the source DFT databases to limit the search by.
    crystalfamily : str or list, optional
        The crystal system families to limit the search by.
    composition : str or list, optional
        The reduced compositions of the structures to limit the search by.
        Element symbols are sorted alphabetically.
    symbols : str or list, optional
        Element symbols in the crystal to limit the search by.
    natoms : int or list, optional
        The number of unique atoms in the crystal's unit cell to limit the
        search by.
    natypes : int or list, optional
        The number(s) of unique atom types to limit the search by.
    local : bool, optional
        Indicates if the local location is to be searched.  Default value
        matches the value set when the database was initialized.
    remote : bool, optional
        Indicates if the remote location is to be searched.  Default value
        matches the value set when the database was initialized.
    prompt : bool, optional
        If prompt=True (default) then a screen input will ask for a selection
        if multiple matching potentials are found.  If prompt=False, then an
        error will be thrown if multiple matches are found.
    format : str, optional
        The file format to save the record in: 'json' or 'xml'.  Default
        is 'json'.
    indent : int, optional
        The number of space indentation spacings to use in the saved
        record for the different tiered levels.  Default is 4.  Giving None
        will create a compact record.
    refresh_cache : bool, optional
        If the local database is of style "local", indicates if the metadata
        cache file is to be refreshed.  If False,
        metadata for new records will be added but the old record metadata
        fields will not be updated.  If True, then the metadata for all
        records will be regenerated, which is needed to update the metadata
        for modified records.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    
    Raises
    ------
    ValueError
        If local or remote is set to True when the corresponding database
        interaction has not been set.
    ValueError
        If multiple or no matching records are discovered.
    """
    self.retrieve_record(
        style='reference_crystal', name=name, dest=dest, local=local,
        remote=remote, prompt=prompt, promptfxn=promptfxn, format=format,
        indent=indent, refresh_cache=refresh_cache, verbose=verbose,
        key=key, id=id, sourcename=sourcename, sourcelink=sourcelink,
        crystalfamily=crystalfamily, composition=composition,
        symbols=symbols, natoms=natoms, natypes=natypes)

def download_reference_crystals(self, 
                                name: Union[str, list, None] = None,
                                key: Union[str, list, None] = None,
                                id: Union[str, list, None] = None,
                                sourcename: Union[str, list, None] = None,
                                sourcelink: Union[str, list, None] = None,
                                crystalfamily: Union[str, list, None] = None,
                                composition: Union[str, list, None] = None,
                                symbols: Union[str, list, None] = None,
                                natoms: Union[int, list, None] = None,
                                natypes: Union[int, list, None] = None,
                                overwrite: bool = False,
                                return_records: bool = False,
                                verbose: bool = False) -> Optional[np.ndarray]:
    """
    Download reference records from the remote and save to localpath.
    
    Parameters
    ----------
    name : str or list, optional
        The record name(s) to parse by.  For reference crystal records, the
        names should correspond to the id.
    id : str or list, optional
        The record id(s) to parse by.  For reference crystal records, the
        id are letters identifying the source database "mp-", "mvc-", or "oqmd-",
        followed by the source database's identification number.
    key : str or list, optional
        UUID4 key(s) to search for.  Each entry has a unique random-generated
        UUID4 key.
    sourcename : str or list, optional
        The full name of source DFT databases to limit the search by.
        "Materials Project" or "Open Quantum Materials Database".            
    sourcelink : str or list, optional
        The web link of the source DFT databases to limit the search by.
    crystalfamily : str or list, optional
        The crystal system families to limit the search by.
    composition : str or list, optional
        The reduced compositions of the structures to limit the search by.
        Element symbols are sorted alphabetically.
    symbols : str or list, optional
        Element symbols in the crystal to limit the search by.
    natoms : int or list, optional
        The number of unique atoms in the crystal's unit cell to limit the
        search by.
    natypes : int or list, optional
        The number(s) of unique atom types to limit the search by.
    overwrite : bool, optional
        Flag indicating if any existing local records with names matching
        remote records are updated (True) or left unchanged (False).  Default
        value is False.
    return_records : bool, optional
        If True, the retrieved record objects are also returned.  Default
        value is False.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    
    Returns
    -------
    numpy.ndarray
        The matching records - returned if return_records=True
    """

    return self.download_records(
        style='relaxed_crystal', name=name, overwrite=overwrite,
        return_records=return_records, verbose=verbose,
        key=key, id=id, sourcename=sourcename, sourcelink=sourcelink,
        crystalfamily=crystalfamily, composition=composition,
        symbols=symbols, natoms=natoms, natypes=natypes)

def fetch_reference_crystal(self,
                            id: str,
                            api_key: Optional[str] = None,
                            local: Optional[bool] = None,
                            remote: Optional[bool] = None, 
                            refresh_cache: bool = False,
                            verbose: bool = False) -> Record:
    """
    Retrieves a single reference crystal.  First, the database is checked
    for matches with the DOI, then with the record name.  If no matches are found
    in the database, then the corresponding crystal structure is downloaded from
    the source database.

    Parameters
    ----------
    id : str
        The reference crystal's unique id.  Combines a database tag "mp-" or
        "oqmd-" and the DFT database's entry id.
    api_key : str, optional
        The user's Materials Project API key or path to a file containing the
        key. Only needed for fetching structures from Materials Project and if
        the key is not set to the "MAPI_KEY" environment variable.
    local : bool, optional
        Indicates if the local location is to be searched.  Default value
        matches the value set when the database was initialized.
    remote : bool, optional
        Indicates if the remote location is to be searched.  Default value
        matches the value set when the database was initialized.
    refresh_cache : bool, optional
        If the local database is of style "local", indicates if the metadata
        cache file is to be refreshed.  If False,
        metadata for new records will be added but the old record metadata
        fields will not be updated.  If True, then the metadata for all
        records will be regenerated, which is needed to update the metadata
        for modified records.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    
    Returns
    -------
    yabadaba.record.Record
        The matching reference crystal record.
    """
    if local is not False or remote is not False:
        # Try fetching based on doi
        try:
            return self.get_reference_crystal(id=id, local=local, remote=remote,
                                              refresh_cache=refresh_cache, verbose=verbose)
        except:
            pass

    # Fetch from source database
    if 'oqmd-' in id:
        record = self.fetch_oqmd_crystal(id)
        if verbose:
            print(f'Crystal retrieved from OQMD')
    else:
        record = self.fetch_mp_crystal(id, api_key=api_key)
        if verbose:
            print(f'Crystal retrieved from Materials Project')

    return record

def fetch_mp_crystals(self,
                      id: str,
                      api_key: Optional[str] = None) -> list:
    """
    Retrieves reference crystals from Materials Project based on id(s).

    Parameters
    ----------
    id : str or list
        The structure id(s) of the crystals to retrieve.
    api_key : str, optional
        The user's Materials Project API key or path to a file containing the key.
        If not given, will use the "MAPI_KEY" environment variable.
    
    Returns
    -------
    list
        All matching reference crystals retrieved.
    """

    # Function-specific imports
    import pymatgen as pmg
    from pymatgen.ext.matproj import MPRester

    # Open connection to Materials Project
    records = []
    with MPRester(api_key) as m:

        # Download missing entries
        try:
            entries = m.query({"material_id": {"$in": aslist(id)}}, ['material_id', 'cif'])
        except:
            raise ValueError('Failed to find Materials Project information')
        else:
            # Convert cif to model and save
            for entry in entries:
                entry_id = entry['material_id']
                struct = pmg.Structure.from_str(entry['cif'], fmt='cif')
                struct = pmg.symmetry.analyzer.SpacegroupAnalyzer(struct).get_conventional_standard_structure()

                # Build record content
                # Build basic record content
                record = load_record('reference_crystal', name=entry_id)
                record.sourcename = "Materials Project"
                record.sourcelink = "https://materialsproject.org/"
                record.ucell = load_pymatgen_Structure(struct).normalize()
                records.append(record)
    
    return records

def fetch_mp_crystal(self,
                     id: str,
                     api_key: Optional[str] = None) -> Record:
    """
    Retrieves a single reference crystal from Materials Project based on id.

    Parameters
    ----------
    id : str
        The structure id of the crystal to retrieve.
    api_key : str, optional
        The user's Materials Project API key. If not given, will use "MAPI_KEY"
        environment variable.
    
    Returns
    -------
    Record
        The retrieved reference crystal.
    """
    # Read api_key from a file
    if Path(api_key).is_file():
        with open(api_key) as f:
            api_key = f.read().strip()

    records = self.fetch_mp_crystals(id, api_key=api_key)
    if len(records) == 1:
        return records[0]
    else:
        raise ValueError('Exactly one record not found?')

def fetch_oqmd_crystal(self, id: str) -> Record:
    """
    Retrieves a single reference crystal from OQMD based on id.

    Parameters
    ----------
    id : str
        The OQMD entry number with "oqmd-" prefix.

    Returns
    -------
    Record
        The retrieved reference crystal.
    """
    
    # Build basic record content
    record = load_record('reference_crystal', name=id)
    record.sourcename = "Open Quantum Materials Database"
    record.sourcelink = "http://oqmd.org/"
    
    # Parse entry number page for structure number
    entry_number = id.replace('oqmd-', '')
    entry_r = requests.get(f'http://oqmd.org/materials/entry/{entry_number}')
    entry_html = entry_r.text        
    start = entry_html.index('href="/materials/structure/') + len('href="/materials/structure/')
    end = start + entry_html[start:].index('"')
    structure_number = entry_html[start:end]

    # Try retrieving poscar of conventional then primitive cells
    try:
        structure_url = f'http://oqmd.org/materials/export/conventional/poscar/{structure_number}'
        structure_r = requests.get(structure_url)
        structure_r.raise_for_status()
    except:
        try:
            structure_url = f'http://oqmd.org/materials/export/primitive/poscar/{structure_number}' 
            structure_r = requests.get(structure_url)
            structure_r.raise_for_status()
        except:
            raise ValueError('Failed to find the poscar file for the structure')
            
    # Load ucell
    record.ucell = load_poscar(structure_r.text).normalize()
    
    return record

def save_reference_crystal(self,
                           crystal: Record,
                           overwrite: bool = False,
                           verbose: bool = False):
    """
    Saves a reference crystal to the local database.
    
    Parameters
    ----------
    crystal : ReferenceCrystal
        The record to save.  
    overwrite : bool, optional
        Indicates what to do when a matching record is found in the local
        location.  If False (default), then the record is not updated.  If
        True, then the record is updated.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    """
    self.save_record(record=crystal, overwrite=overwrite, verbose=verbose)


# coding: utf-8

# Standard Python libraries
from typing import Union

def boolean(value: Union[str, bool]) -> bool:
    """
    Allows conversion of strings to Booleans.
    
    Parameters
    ----------
    value : str or bool
        If str, then 'true' and 't' become True and 'false' and 'f' become
        false. If bool, simply return the value.
        
    Returns
    -------
    bool
        Equivalent bool of value.
        
    Raises
    ------
    ValueError
        If value is unrecognized.
    """
    
    # Pass Boolean values through without changing
    if value is True:
        return True
    elif value is False:
        return False
    
    # Convert strings
    elif value.lower() in ['true', 't']:
        return True
    elif value.lower() in ['false', 'f']:
        return False
    
    # Issue error for invalid string
    else:
        raise ValueError('Invalid Boolean string')


# coding: utf-8

# Standard Python imports
import io
from typing import Optional, Union, Tuple
import uuid

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

# https://github.com/usnistgov/yabadaba
from yabadaba.record import Record
from yabadaba import load_query

# https://pandas.pydata.org/
import pandas as pd

# atomman imports
from ... import System

class ReferenceCrystal(Record):
    """
    Class for representing reference_crystal records that provide the structure
    information for DFT relaxed crystal structures obtained from DFT databases.
    """
    @property
    def style(self) -> str:
        """str: The record style"""
        return 'reference_crystal'

    @property
    def modelroot(self) -> str:
        """str: The root element of the content"""
        return 'reference-crystal'

    @property
    def xsd_filename(self) -> Tuple[str, str]:
        """tuple: The module path and file name of the record's xsd schema"""
        return ('atomman.library.xsd', f'{self.style}.xsd')

    @property
    def id(self) -> str:
        """str : The unique id for the record"""
        return self.__id

    @id.setter
    def id(self, value: str):
        if value is None:
            self.__id = None
        else:
            self.__id = str(value)

    @property
    def key(self) -> str:
        """str : A UUID4 key assigned to the record"""
        return self.__key

    @key.setter
    def key(self, value: str):
        if value is None:
            self.__key = str(uuid.uuid4())
        else:
            self.__key = str(value)

    @property
    def sourcename(self) -> str:
        """str : Name of the crystal's source database"""
        return self.__sourcename

    @sourcename.setter
    def sourcename(self, value: str):
        if value is None:
            self.__sourcename = None
        else:
            self.__sourcename = str(value)

    @property
    def sourcelink(self) -> str:
        """str : URL for the crystal's source database"""
        return self.__sourcelink    

    @sourcelink.setter
    def sourcelink(self, value: str):
        if value is None:
            self.__sourcelink = None
        else:
            self.__sourcelink = str(value)

    @property
    def ucell(self) -> System:
        """atomman.System : The unit cell system for the crystal"""
        if self.__ucell is None:
            raise ValueError('ucell information not set')
        elif not isinstance(self.__ucell, System):
            self.__ucell = System(model=self.__ucell)
        return self.__ucell

    @ucell.setter
    def ucell(self, value: System):
        if isinstance(value, System):       
            self.__ucell = value
        else:
            raise TypeError('ucell must be an atomman.System')

    @property
    def composition(self) -> str:
        """str : The crystal's composition"""
        if self.__composition is None:
            self.__composition = self.ucell.composition
        return self.__composition

    @property
    def symbols(self) -> list:
        """list : The list of element model symbols"""
        if self.__symbols is None:
            self.__symbols = self.ucell.symbols
        return self.__symbols

    @property
    def natoms(self) -> int:
        """int : The number of atoms in the unit cell"""
        if self.__natoms is None:
            self.__natoms = self.ucell.natoms
        return self.__natoms

    @property
    def natypes(self) -> int:
        """int : The number of atom types in the unit cell"""
        if self.__natypes is None:
            self.__natypes = self.ucell.natypes
        return self.__natypes

    @property
    def crystalfamily(self) -> str:
        """str : The crystal's system family"""
        if self.__crystalfamily is None:
            self.__crystalfamily = self.ucell.box.identifyfamily()
        return self.__crystalfamily

    @property
    def a(self) -> float:
        """float : The unit cell's a lattice parameter"""
        if self.__a is None:
            self.__a = self.ucell.box.a
        return self.__a

    @property
    def b(self) -> float:
        """float : The unit cell's b lattice parameter"""
        if self.__b is None:
            self.__b = self.ucell.box.b
        return self.__b

    @property
    def c(self) -> float:
        """float : The unit cell's c lattice parameter"""
        if self.__c is None:
            self.__c = self.ucell.box.c
        return self.__c

    @property
    def alpha(self) -> float:
        """float : The unit cell's alpha lattice angle"""
        if self.__alpha is None:
            self.__alpha = self.ucell.box.alpha
        return self.__alpha

    @property
    def beta(self) -> float:
        """float : The unit cell's beta lattice angle"""
        if self.__beta is None:
            self.__beta = self.ucell.box.beta
        return self.__beta

    @property
    def gamma(self) -> float:
        """float : The unit cell's gamma lattice angle"""
        if self.__gamma is None:
            self.__gamma = self.ucell.box.gamma
        return self.__gamma

    def set_values(self,
                   name: Optional[str] = None,
                   id: Optional[str] = None,
                   key: Optional[str] = None,
                   sourcename: Optional[str] = None,
                   sourcelink: Optional[str] = None,
                   ucell: Optional[System] = None):
        """
        Sets multiple object values.

        Parameters
        ----------
        name : str, optional
            The name to use for saving the record.  Either name or id should
            be given as they are treated as aliases for this record style.
        id : str, optional
            The unique identifier for the record.  Should be composed of a
            source database tag plus the source database's unique identifier.
        key : str, optional
            A UUID4 key assigned to the record.  Note that if not given a new
            random key will be assigned and therefore the keys might not match
            if similar records were generated independently.
        sourcename : str, optional
            The name of the source database where the reference record was
            retrieved.
        sourcelink : str, optional
            The URL to the source database where the reference record was
            retrieved.
        ucell : atomman.System, optional
            A small unit cell system associated with the reference crystal.
        """
        
        if name is None and id is not None:
            self.name = id
            self.id = id
        elif name is not None and id is None:
            self.name = name
            self.id = name
        else:
            self.name = name
            self.id = id

        self.key = key
        self.sourcename = sourcename
        self.sourcelink = sourcelink
        if ucell is not None:
            self.ucell = ucell
        
        self.__symbols = None
        self.__composition = None
        self.__crystalfamily = None
        self.__natypes = None
        self.__a = None
        self.__b = None
        self.__c = None
        self.__alpha = None
        self.__beta = None
        self.__gamma = None

    def build_model(self) -> DM:
        """
        Returns the object info as data model content
        
        Returns
        ----------
        DataModelDict
            The data model content.
        """

        refmodel = DM()
        
        refmodel['key'] = self.key
        refmodel['id'] = self.id

        refmodel['source'] = DM()
        refmodel['source']['name'] = self.sourcename
        refmodel['source']['link'] = self.sourcelink
        
        refmodel['system-info'] = DM()
        symbols = self.symbols
        if len(symbols) == 1:
            refmodel['system-info']['symbol'] = self.symbols[0]
        else:
            refmodel['system-info']['symbol'] = self.symbols
        refmodel['system-info']['composition'] = self.composition
        
        refmodel['system-info']['cell'] = DM()
        refmodel['system-info']['cell']['crystal-family'] = self.crystalfamily
        refmodel['system-info']['cell']['natypes'] = self.natypes
        refmodel['system-info']['cell']['a'] = self.a
        refmodel['system-info']['cell']['b'] = self.b
        refmodel['system-info']['cell']['c'] = self.c
        refmodel['system-info']['cell']['alpha'] = self.alpha
        refmodel['system-info']['cell']['beta'] = self.beta
        refmodel['system-info']['cell']['gamma'] = self.gamma
        
        refmodel['atomic-system'] = self.ucell.model()['atomic-system']

        model = DM([('reference-crystal', refmodel)])
        self._set_model(model)
        return model
        
    def load_model(self,
                   model: Union[str, io.IOBase, DM],
                   name: Optional[str] = None):
        """
        Loads record contents from a given model.

        Parameters
        ----------
        model : str or DataModelDict
            The model contents of the record to load.
        name : str, optional
            The name to assign to the record.  Often inferred from other
            attributes if not given.
        """
        super().load_model(model, name=name)        
        crystal = self.model[self.modelroot]
        
        self.__key = crystal['key']
        self.__id = crystal['id']
        self.__sourcename = crystal['source']['name']
        self.__sourcelink = crystal['source']['link']
        
        self.__symbols = crystal['system-info'].aslist('symbol')
        self.__composition = crystal['system-info']['composition']
        self.__crystalfamily = crystal['system-info']['cell']['crystal-family']
        self.__natypes = crystal['system-info']['cell']['natypes']
        self.__a = crystal['system-info']['cell']['a']
        self.__b = crystal['system-info']['cell']['b']
        self.__c = crystal['system-info']['cell']['c']
        self.__alpha = crystal['system-info']['cell']['alpha']
        self.__beta = crystal['system-info']['cell']['beta']
        self.__gamma = crystal['system-info']['cell']['gamma']

        self.__natoms = crystal['atomic-system']['atoms']['natoms']

        self.__ucell = crystal

        # Set name as id if no name given
        try:
            self.name
        except:
            self.name = self.id

    def metadata(self) -> dict:
        """
        Generates a dict of simple metadata values associated with the record.
        Useful for quickly comparing records and for building pandas.DataFrames
        for multiple records of the same style.
        """
        params = {}
        params['name'] = self.name
        params['key'] = self.key
        params['id'] = self.id
        params['sourcename'] = self.sourcename
        params['sourcelink'] = self.sourcelink

        params['crystalfamily'] = self.crystalfamily
        params['natypes'] = self.natypes
        params['symbols'] = self.symbols
        params['composition'] = self.composition

        params['a'] = self.a
        params['b'] = self.b
        params['c'] = self.c
        params['alpha'] = self.alpha
        params['beta'] = self.beta
        params['gamma'] = self.gamma
        params['natoms'] = self.natoms

        return params

    @property
    def queries(self) -> dict:
        """dict: Query objects and their associated parameter names."""
        return {
            'key': load_query(
                style='str_match',
                name='key', 
                path=f'{self.modelroot}.key'),
            'id': load_query(
                style='str_match',
                name='id',
                path=f'{self.modelroot}.id'),
            'sourcename': load_query(
                style='str_match',
                name='sourcename',
                path=f'{self.modelroot}.source.name'),
            'sourcelink': load_query(
                style='str_match',
                name='sourcelink',
                path=f'{self.modelroot}.source.link'),
            'crystalfamily': load_query(
                style='str_match',
                name='crystalfamily',
                path=f'{self.modelroot}.system-info.cell.crystal-family'),
            'composition': load_query(
                style='str_match',
                name='composition',
                path=f'{self.modelroot}.system-info.composition'),
            'symbols': load_query(
                style='list_contains',
                name='symbols',
                path=f'{self.modelroot}.system-info.symbol'),
            'natoms': load_query(
                style='int_match',
                name='natoms',
                path=f'{self.modelroot}.atomic-system.atoms.natoms'),
            'natypes': load_query(
                style='int_match',
                name='natypes',
                path=f'{self.modelroot}.system-info.cell.natypes'),
        }

    def pandasfilter(self,
                     dataframe: pd.DataFrame,
                     name: Union[str, list, None] = None,
                     id: Union[str, list, None] = None,
                     key: Union[str, list, None] = None,
                     sourcename: Union[str, list, None] = None,
                     sourcelink: Union[str, list, None] = None,
                     crystalfamily: Union[str, list, None] = None,
                     composition: Union[str, list, None] = None,
                     symbols: Union[str, list, None] = None,
                     natoms: Union[int, list, None] = None,
                     natypes: Union[int, list, None] = None) -> pd.Series:
        """
        Filters a pandas.DataFrame based on kwargs values for the record style.
        
        Parameters
        ----------
        dataframe : pandas.DataFrame
            A table of metadata for multiple records of the record style.
        name : str or list
            The record name(s) to parse by.
        id : str or list
            The record id(s) to parse by.
        key : str or list
            The record key(s) to parse by.
        sourcename : str or list
            The name(s) of source databases to parse by.
        sourcelink : str or list
            The URL(s) of source databases to parse by.
        crystalfamily : str or list
            Crystal structure families to parse by.
        composition : str or list
            Compositions to parse by.
        symbols : str or list
            Element model symbol(s) to parse by.
        natoms : int or list
            Number of atoms in the unit cell to parse by.
        natypes : int or list
            Number of atom types to parse by.
        
        Returns
        -------
        pandas.Series
            Boolean map of matching values
        """
        matches = super().pandasfilter(dataframe, name=name, id=id, key=key,
                                       sourcename=sourcename, sourcelink=sourcelink,
                                       crystalfamily=crystalfamily,
                                       composition=composition, symbols=symbols,
                                       natoms=natoms, natypes=natypes)
        return matches

    def mongoquery(self,
                   name: Union[str, list, None] = None,
                   id: Union[str, list, None] = None,
                   key: Union[str, list, None] = None,
                   sourcename: Union[str, list, None] = None,
                   sourcelink: Union[str, list, None] = None,
                   crystalfamily: Union[str, list, None] = None,
                   composition: Union[str, list, None] = None,
                   symbols: Union[str, list, None] = None,
                   natoms: Union[int, list, None] = None,
                   natypes: Union[int, list, None] = None) -> dict:
        """
        Builds a Mongo-style query based on kwargs values for the record style.
        
        Parameters
        ----------
        name : str or list
            The record name(s) to parse by.
        id : str or list
            The record id(s) to parse by.
        key : str or list
            The record key(s) to parse by.
        sourcename : str or list
            The name(s) of source databases to parse by.
        sourcelink : str or list
            The URL(s) of source databases to parse by.
        crystalfamily : str or list
            Crystal structure families to parse by.
        composition : str or list
            Compositions to parse by.
        symbols : str or list
            Element model symbol(s) to parse by.
        natoms : int or list
            Number of atoms in the unit cell to parse by.
        natypes : int or list
            Number of atom types to parse by.
        
        Returns
        -------
        dict
            The Mongo-style query
        """     
        mquery = super().mongoquery(name=name, id=id, key=key,
                                    sourcename=sourcename, sourcelink=sourcelink,
                                    crystalfamily=crystalfamily,
                                    composition=composition, symbols=symbols,
                                    natoms=natoms, natypes=natypes)
        return mquery

    def cdcsquery(self,
                  id: Union[str, list, None] = None,
                  key: Union[str, list, None] = None,
                  sourcename: Union[str, list, None] = None,
                  sourcelink: Union[str, list, None] = None,
                  crystalfamily: Union[str, list, None] = None,
                  composition: Union[str, list, None] = None,
                  symbols: Union[str, list, None] = None,
                  natoms: Union[int, list, None] = None,
                  natypes: Union[int, list, None] = None) -> dict:
        """
        Builds a CDCS-style query based on kwargs values for the record style.
        
        Parameters
        ----------
        id : str or list
            The record id(s) to parse by.
        key : str or list
            The record key(s) to parse by.
        sourcename : str or list
            The name(s) of source databases to parse by.
        sourcelink : str or list
            The URL(s) of source databases to parse by.
        crystalfamily : str or list
            Crystal structure families to parse by.
        composition : str or list
            Compositions to parse by.
        symbols : str or list
            Element model symbol(s) to parse by.
        natoms : int or list
            Number of atoms in the unit cell to parse by.
        natypes : int or list
            Number of atom types to parse by.
        
        Returns
        -------
        dict
            The CDCS-style query
        """
        mquery = super().cdcsquery(id=id, key=key,
                                    sourcename=sourcename, sourcelink=sourcelink,
                                    crystalfamily=crystalfamily,
                                    composition=composition, symbols=symbols,
                                    natoms=natoms, natypes=natypes)
        return mquery


# coding: utf-8

# Standard Python imports
import io
from typing import Optional, Union, Tuple

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

# https://github.com/usnistgov/yabadaba
from yabadaba.record import Record
from yabadaba import load_query

# https://pandas.pydata.org/
import pandas as pd

# atomman imports
import atomman.unitconvert as uc
from ... import System

class RelaxedCrystal(Record):
    """
    Class for representing relaxed_crystal records that provide the structure
    information for crystal structures relaxed using a specific interatomic
    potential.
    """
    def __init__(self,
                 model: Union[str, io.IOBase, DM, None] = None,
                 name: Optional[str] = None):
        """
        Initializes a Record object for a given style.
        
        Parameters
        ----------
        model : str, file-like object, DataModelDict
            The contents of the record.
        name : str, optional
            The unique name to assign to the record.  If model is a file
            path, then the default record name is the file name without
            extension.
        """
        if model is not None:
            super().__init__(model=model, name=name)
        elif name is not None:
            self.name = name

    @property
    def style(self) -> str:
        """str: The record style"""
        return 'relaxed_crystal'

    @property
    def modelroot(self) -> str:
        """str: The root element of the content"""
        return 'relaxed-crystal'

    @property
    def xsd_filename(self) -> Tuple[str, str]:
        """tuple: The module path and file name of the record's xsd schema"""
        return ('atomman.library.xsd', f'{self.style}.xsd')

    def load_model(self,
                   model: Union[str, io.IOBase, DM],
                   name: Optional[str] = None):
        """
        Loads record contents from a given model.

        Parameters
        ----------
        model : str or DataModelDict
            The model contents of the record to load.
        name : str, optional
            The name to assign to the record.  Often inferred from other
            attributes if not given.
        """
        super().load_model(model, name=name)        
        crystal = self.model[self.modelroot]
        
        self.__key = crystal['key']
        self.__method = crystal['method']
        self.__standing = crystal['standing']

        self.__potential_LAMMPS_id = crystal['potential-LAMMPS']['id']
        self.__potential_LAMMPS_key = crystal['potential-LAMMPS']['key']
        self.__potential_id = crystal['potential-LAMMPS']['potential']['id']
        self.__potential_key = crystal['potential-LAMMPS']['potential']['key']

        self.__family = crystal['system-info']['family']
        self.__parent_key = crystal['system-info']['parent_key']
        self.__symbols = crystal['system-info'].aslist('symbol')
        self.__composition = crystal['system-info']['composition']
        self.__crystalfamily = crystal['system-info']['cell']['crystal-family']
        self.__natypes = crystal['system-info']['cell']['natypes']
        self.__a = crystal['system-info']['cell']['a']
        self.__b = crystal['system-info']['cell']['b']
        self.__c = crystal['system-info']['cell']['c']
        self.__alpha = crystal['system-info']['cell']['alpha']
        self.__beta = crystal['system-info']['cell']['beta']
        self.__gamma = crystal['system-info']['cell']['gamma']
        
        self.__natoms = crystal['atomic-system']['atoms']['natoms']

        self.__potential_energy = uc.value_unit(crystal['potential-energy'])
        self.__cohesive_energy = uc.value_unit(crystal['cohesive-energy'])
        
        self.__ucell = None

        # Set name as key if no name given
        try:
            self.name
        except:
            self.name = self.key

    @property
    def key(self) -> str:
        """str : A UUID4 key assigned to the record"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__key

    @property
    def method(self) -> str:
        """str : Indicates the relaxation method used: box, static or dynamic"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__method
    
    @property
    def standing(self) -> str:
        """str : 'good' or 'bad', with bad indicating it to be a duplicate or transformation"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__standing
    
    @property
    def family(self) -> str:
        """str : The associated prototype/reference crystal id that the relaxed crystal is based on"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__family

    @property
    def parent_key(self) -> str:
        """str : The key assigned to the record of the relaxation calculation used"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__parent_key

    @property
    def potential_LAMMPS_id(self) -> str:
        """str : The id of the LAMMPS implementation used to relax the crystal"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__potential_LAMMPS_id

    @property
    def potential_LAMMPS_key(self) -> str:
        """str : The key of the LAMMPS implementation used to relax the crystal"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__potential_LAMMPS_key

    @property
    def potential_id(self) -> str:
        """str : The id of the potential model used to relax the crystal"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__potential_id

    @property
    def potential_key(self) -> str:
        """str : The key of the potential model used to relax the crystal"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__potential_key

    @property
    def cohesive_energy(self) -> float:
        """float : The computed per-atom cohesive energy"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__cohesive_energy

    @property
    def potential_energy(self) -> float:
        """float : The measured per-atom potential energy"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__potential_energy

    @property
    def composition(self) -> str:
        """str : The crystal's composition"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__composition

    @property
    def symbols(self) -> list:
        """list : The list of element model symbols"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__symbols

    @property
    def natoms(self) -> int:
        """int : The number of atoms in the unit cell"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__natoms

    @property
    def natypes(self) -> int:
        """int : The number of atom types in the unit cell"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__natypes

    @property
    def crystalfamily(self) -> str:
        """str : The crystal's system family"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__crystalfamily

    @property
    def a(self) -> float:
        """float : The unit cell's a lattice parameter"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__a

    @property
    def b(self) -> float:
        """float : The unit cell's b lattice parameter"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__b

    @property
    def c(self) -> float:
        """float : The unit cell's c lattice parameter"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__c

    @property
    def alpha(self) -> float:
        """float : The unit cell's alpha lattice angle"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__alpha

    @property
    def beta(self) -> float:
        """float : The unit cell's beta lattice angle"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__beta

    @property
    def gamma(self) -> float:
        """float : The unit cell's gamma lattice angle"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__gamma

    @property
    def ucell(self) -> System:
        """atomman.System : The unit cell system for the crystal"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        if self.__ucell is None:
            self.__ucell = System(model=self.model)
        return self.__ucell

    def build_model(self) -> DM:
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.model

    def metadata(self) -> dict:
        """
        Generates a dict of simple metadata values associated with the record.
        Useful for quickly comparing records and for building pandas.DataFrames
        for multiple records of the same style.
        """
        params = {}
        params['name'] = self.name
        params['key'] = self.key
        params['method'] = self.method
        params['standing'] = self.standing
        params['family'] = self.family
        params['parent_key'] = self.parent_key
        
        params['potential_LAMMPS_id'] = self.potential_LAMMPS_id
        params['potential_LAMMPS_key'] = self.potential_LAMMPS_key
        params['potential_id'] = self.potential_id
        params['potential_key'] = self.potential_key
        
        params['crystalfamily'] = self.crystalfamily
        params['natypes'] = self.natypes
        params['symbols'] = self.symbols
        params['composition'] = self.composition

        params['a'] = self.a
        params['b'] = self.b
        params['c'] = self.c
        params['alpha'] = self.alpha
        params['beta'] = self.beta
        params['gamma'] = self.gamma
        params['natoms'] = self.natoms

        params['potential_energy'] = self.potential_energy
        params['cohesive_energy'] = self.cohesive_energy

        return params

    @property
    def queries(self) -> dict:
        """dict: Query objects and their associated parameter names."""
        return {
            'key': load_query(
                style='str_match',
                name='key', 
                path=f'{self.modelroot}.key'),
            'method': load_query(
                style='str_match',
                name='method',
                path=f'{self.modelroot}.method'),
            'standing': load_query(
                style='str_match',
                name='standing',
                path=f'{self.modelroot}.standing'),
            'family': load_query(
                style='str_match',
                name='family',
                path=f'{self.modelroot}.system-info.family'),
            'parent_key': load_query(
                style='str_match',
                name='parent_key',
                path=f'{self.modelroot}.system-info.parent_key'),
            'potential_LAMMPS_id': load_query(
                style='str_match',
                name='potential_LAMMPS_id',
                path=f'{self.modelroot}.potential-LAMMPS.id'),
            'potential_LAMMPS_key': load_query(
                style='str_match',
                name='potential_LAMMPS_key',
                path=f'{self.modelroot}.potential-LAMMPS.key'),
            'potential_id': load_query(
                style='str_match',
                name='potential_id',
                path=f'{self.modelroot}.potential-LAMMPS.potential.id'),
            'potential_key': load_query(
                style='str_match',
                name='potential_key',
                path=f'{self.modelroot}.potential-LAMMPS.potential.key'),
            'crystalfamily': load_query(
                style='str_match',
                name='crystalfamily',
                path=f'{self.modelroot}.system-info.cell.crystal-family'),
            'composition': load_query(
                style='str_match',
                name='composition',
                path=f'{self.modelroot}.system-info.composition'),
            'symbols': load_query(
                style='list_contains',
                name='symbols',
                path=f'{self.modelroot}.system-info.symbol'),
            'natoms': load_query(
                style='int_match',
                name='natoms',
                path=f'{self.modelroot}.atomic-system.atoms.natoms'),
            'natypes': load_query(
                style='int_match',
                name='natypes',
                path=f'{self.modelroot}.system-info.cell.natypes'),
        }

    def pandasfilter(self,
                     dataframe: pd.DataFrame,
                     name: Union[str, list, None] = None,
                     key: Union[str, list, None] = None,
                     method: Union[str, list, None] = None,
                     standing: Union[str, list, None] = None,
                     family: Union[str, list, None] = None,
                     parent_key: Union[str, list, None] = None, 
                     potential_LAMMPS_id: Union[str, list, None] = None,
                     potential_LAMMPS_key: Union[str, list, None] = None,
                     potential_id: Union[str, list, None] = None,
                     potential_key: Union[str, list, None] = None,
                     crystalfamily: Union[str, list, None] = None,
                     composition: Union[str, list, None] = None,
                     symbols: Union[str, list, None] = None,
                     natoms: Union[int, list, None] = None,
                     natypes: Union[int, list, None] = None) -> pd.Series:
        """
        Filters a pandas.DataFrame based on kwargs values for the record style.
        
        Parameters
        ----------
        dataframe : pandas.DataFrame
            A table of metadata for multiple records of the record style.
        name : str or list
            The record name(s) to parse by.
        key : str or list
            The record key(s) to parse by.
        method : str, None or list
            The relaxation method(s) to parse by.  Default value is 'dynamic'.
            Set to None to access records for all relaxation methods.
        standing : str, None or list
            The standing status to parse by.  Default value is 'good'  Set to
            None to access records for all standings.
        family : str or list
            The prototype/reference id(s) that the relaxed crystals are based
            on to parse by.
        parent_key : str or list
            The key(s) of the relaxation calculation records to parse by.
        potential_LAMMPS_id : str or list
            LAMMPS potential implementation id(s) to parse by.
        potential_LAMMPS_key : str or list
            LAMMPS potential implementation keys(s) to parse by.
        potential_id : str or list
            Potential id(s) to parse by.
        potential_key : str or list
            Potential key(s) to parse by.
        crystalfamily : str or list
            Crystal structure families to parse by.
        composition : str or list
            Compositions to parse by.
        symbols : str or list
            Element model symbol(s) to parse by.
        natoms : int or list
            Number of atoms in the unit cell to parse by.
        natypes : int or list
            Number of atom types to parse by.
        
        Returns
        -------
        pandas.Series, numpy.NDArray
            Boolean map of matching values
        """
        matches = super().pandasfilter(dataframe, name=name, key=key,
                                       method=method, standing=standing,
                                       family=family, parent_key=parent_key, 
                                       potential_LAMMPS_id=potential_LAMMPS_id,
                                       potential_LAMMPS_key=potential_LAMMPS_key,
                                       potential_id=potential_id,
                                       potential_key=potential_key,
                                       crystalfamily=crystalfamily,
                                       composition=composition, symbols=symbols,
                                       natoms=natoms, natypes=natypes)
        return matches

    def mongoquery(self,
                   name: Union[str, list, None] = None,
                   key: Union[str, list, None] = None,
                   method: Union[str, list, None] = None,
                   standing: Union[str, list, None] = None,
                   family: Union[str, list, None] = None,
                   parent_key: Union[str, list, None] = None, 
                   potential_LAMMPS_id: Union[str, list, None] = None,
                   potential_LAMMPS_key: Union[str, list, None] = None,
                   potential_id: Union[str, list, None] = None,
                   potential_key: Union[str, list, None] = None,
                   crystalfamily: Union[str, list, None] = None,
                   composition: Union[str, list, None] = None,
                   symbols: Union[str, list, None] = None,
                   natoms: Union[int, list, None] = None,
                   natypes: Union[int, list, None] = None) -> dict:
        """
        Builds a Mongo-style query based on kwargs values for the record style.
        
        Parameters
        ----------
        name : str or list
            The record name(s) to parse by.
        key : str or list
            The record key(s) to parse by.
        method : str, None or list
            The relaxation method(s) to parse by.  Default value is 'dynamic'.
            Set to None to access records for all relaxation methods.
        standing : str, None or list
            The standing status to parse by.  Default value is 'good'  Set to
            None to access records for all standings.
        family : str or list
            The prototype/reference id(s) that the relaxed crystals are based
            on to parse by.
        parent_key : str or list
            The key(s) of the relaxation calculation records to parse by.
        potential_LAMMPS_id : str or list
            LAMMPS potential implementation id(s) to parse by.
        potential_LAMMPS_key : str or list
            LAMMPS potential implementation keys(s) to parse by.
        potential_id : str or list
            Potential id(s) to parse by.
        potential_key : str or list
            Potential key(s) to parse by.
        crystalfamily : str or list
            Crystal structure families to parse by.
        composition : str or list
            Compositions to parse by.
        symbols : str or list
            Element model symbol(s) to parse by.
        natoms : int or list
            Number of atoms in the unit cell to parse by.
        natypes : int or list
            Number of atom types to parse by.
        
        Returns
        -------
        dict
            The Mongo-style query
        """     
        mquery = super().mongoquery(name=name, key=key,
                                    method=method, standing=standing,
                                    family=family, parent_key=parent_key, 
                                    potential_LAMMPS_id=potential_LAMMPS_id,
                                    potential_LAMMPS_key=potential_LAMMPS_key,
                                    potential_id=potential_id,
                                    potential_key=potential_key,
                                    crystalfamily=crystalfamily,
                                    composition=composition, symbols=symbols,
                                    natoms=natoms, natypes=natypes)
        return mquery

    def cdcsquery(self,
                  key: Union[str, list, None] = None,
                  method: Union[str, list, None] = None,
                  standing: Union[str, list, None] = None,
                  family: Union[str, list, None] = None,
                  parent_key: Union[str, list, None] = None, 
                  potential_LAMMPS_id: Union[str, list, None] = None,
                  potential_LAMMPS_key: Union[str, list, None] = None,
                  potential_id: Union[str, list, None] = None,
                  potential_key: Union[str, list, None] = None,
                  crystalfamily: Union[str, list, None] = None,
                  composition: Union[str, list, None] = None,
                  symbols: Union[str, list, None] = None,
                  natoms: Union[int, list, None] = None,
                  natypes: Union[int, list, None] = None) -> dict:
        """
        Builds a CDCS-style query based on kwargs values for the record style.
        
        Parameters
        ----------
        key : str or list
            The record key(s) to parse by.
        method : str, None or list
            The relaxation method(s) to parse by.  Default value is 'dynamic'.
            Set to None to access records for all relaxation methods.
        standing : str, None or list
            The standing status to parse by.  Default value is 'good'  Set to
            None to access records for all standings.
        family : str or list
            The prototype/reference id(s) that the relaxed crystals are based
            on to parse by.
        parent_key : str or list
            The key(s) of the relaxation calculation records to parse by.
        potential_LAMMPS_id : str or list
            LAMMPS potential implementation id(s) to parse by.
        potential_LAMMPS_key : str or list
            LAMMPS potential implementation keys(s) to parse by.
        potential_id : str or list
            Potential id(s) to parse by.
        potential_key : str or list
            Potential key(s) to parse by.
        crystalfamily : str or list
            Crystal structure families to parse by.
        composition : str or list
            Compositions to parse by.
        symbols : str or list
            Element model symbol(s) to parse by.
        natoms : int or list
            Number of atoms in the unit cell to parse by.
        natypes : int or list
            Number of atom types to parse by.
        
        Returns
        -------
        dict
            The CDCS-style query
        """
        mquery = super().cdcsquery(key=key,
                                    method=method, standing=standing,
                                    family=family, parent_key=parent_key, 
                                    potential_LAMMPS_id=potential_LAMMPS_id,
                                    potential_LAMMPS_key=potential_LAMMPS_key,
                                    potential_id=potential_id,
                                    potential_key=potential_key,
                                    crystalfamily=crystalfamily,
                                    composition=composition, symbols=symbols,
                                    natoms=natoms, natypes=natypes)
        return mquery


# coding: utf-8

# Standard Python libraries
from copy import deepcopy
from typing import Optional

# atomman imports
from ...lammps import style
from ...tools import indexstr

def process_prop_info(prop_name: Optional[list] = None,
                      table_name: Optional[list] = None,
                      shape: Optional[list] = None,
                      unit: Optional[list] = None,
                      dtype: Optional[list] = None,
                      prop_info: Optional[list] = None,
                      lammps_units: str = 'metal') -> list:
    """
    Handles common setting of prop_info terms
    
    Parameters
    ----------
    prop_name : list, optional
        The Atoms properties to include.  If neither prop_name or prop_info are
        given, all system properties will be included.
    table_name : list, optional
        The dump table column name(s) that correspond to each prop_name.  If not
        given, the table_name values will be based on the prop_name and shape
        values.
    shape : list, optional
        The shape of each per-atom property.  If not given, will be inferred
        from the length of each table_name value.
    unit : list, optional
        Lists the units for each prop_name as stored in the table.  For a
        value of None, no conversion will be performed for that property.  For
        a value of 'scaled', the corresponding table values will be taken in
        box-scaled units.  If not given, unit values will be taken based on
        lammps_units if prop_name corresponds to a standard LAMMPS property,
        otherwise will be set to None (no conversion).
    dtype : list, optional
        Allows for the data type of each property to be explicitly given.
        Values of None will infer the data type from the corresponding
        property values.  If not given, all values will be None.
    prop_info : list of dict, optional
        Structured form of property conversion parameters, in which each
        dictionary in the list corresponds to a single atoms property.  Each
        dictionary must have a 'prop_name' field, and can optionally have
        'table_name', 'shape', 'unit', and 'dtype' fields.
    lammps_units : str, optional
        The LAMMPS units option associated with the table values.  This is used
        for the box dimensions and default units for standard dump properties
        (not compute/fix definitions).

    Returns
    -------
    prop_info : list of dict
        The filled-in prop_info structure. Only returned if
        return_prop_info is True.
    """
    
    if prop_info is not None:
        # Check that competing parameters are not given
        try:
            assert prop_name is None
            assert table_name is None
            assert shape is None
            assert unit is None
            assert dtype is None
        except AssertionError as e:
            raise ValueError('prop_info cannot be given with prop_name, table_name, shape, unit, or dtype') from e
        prop_info = deepcopy(prop_info)
    
    # Combine list parameters into prop_info
    else:
        
        # Check that prop_name is given
        if prop_name is None:
            raise ValueError('prop_info or prop_name is required.')
        numprops = len(prop_name)
        
        # Check that sizes match
        try:
            assert table_name is None or len(table_name) == numprops
            assert shape is None or len(shape) == numprops
            assert unit is None or len(unit) == numprops
            assert dtype is None or len(dtype) == numprops
        except AssertionError as e:
            raise ValueError('any of prop_name, table_name, shape, unit, and dtype given must be the same length') from e
        
        # Build prop_info
        prop_info = []
        for i in range(len(prop_name)):
            
            # Search for prop_name match
            for sinfo in standard_conversions(lammps_units):
                match = False
                if sinfo['prop_name'] == prop_name[i]:
                    match = True
                    break
            
            pinfo = {}
            pinfo['prop_name'] = prop_name[i]
            
            if table_name is not None:
                pinfo['table_name'] = table_name[i]
            elif match is True and 'table_name' in sinfo:
                pinfo['table_name'] = sinfo['table_name']
            
            if shape is not None:
                pinfo['shape'] = shape[i]
            elif match is True and 'shape' in sinfo:
                pinfo['shape'] = sinfo['shape']
                
            if unit is not None:
                pinfo['unit'] = unit[i]
            elif match is True and 'unit' in sinfo:
                pinfo['unit'] = sinfo['unit']
                
            if dtype is not None:
                pinfo['dtype'] = dtype[i]
            elif match is True and 'dtype' in sinfo:
                pinfo['dtype'] = sinfo['dtype']
            
            prop_info.append(pinfo)
    
    # Fill in missing default values
    for prop in prop_info:
        
        # prop_name is required
        if 'prop_name' not in prop:
            raise KeyError('prop_name required for each entry in prop_info')
        
        # Generate table_name from prop_name and shape
        if 'table_name' not in prop:
            prop['shape'] = prop.get('shape', ())
            prop['table_name'] = []
            for index, istr in indexstr(prop['shape']): #pylint: disable=unused-variable
                prop['table_name'].append(prop['prop_name'] + istr)
        # Make certain table_name is a list
        else:
            try:
                assert not isinstance(prop['table_name'], str)
                prop['table_name'] = list(prop['table_name'])
            except:
                prop['table_name'] = [prop['table_name']]
        
        # Take shape as length of table_name
        if 'shape' not in prop:
            numtnames = len(prop['table_name'])
            if numtnames == 1:
                prop['shape'] = ()
            else:
                prop['shape'] = (numtnames, )
        # Make certain shape is a tuple
        else:
            try:
                prop['shape'] = tuple(prop['shape'])
            except:
                prop['shape'] = (prop['shape'],)
        
        # Default unit and dtype values are None
        prop['unit'] = prop.get('unit', None)
        prop['dtype'] = prop.get('dtype', None)
    
    return prop_info

def standard_conversions(lammps_units: str = 'metal') -> list:
    """
    Gives pre-defined conversion information for standard LAMMPS variables.
    """
    
    lammps_unit = style.unit(lammps_units)
    
    return [{"prop_name": "atom_id",
             "table_name": "id"},
            
            {"prop_name": "m_id",
             "table_name": "mol"},
            
            {"prop_name": "p_id",
             "table_name": "proc"},
            
            {"prop_name": "p_id_plus1",
             "table_name": "procp1"},
            
            {"prop_name": "atype",
             "table_name": "type"},
            
            {"prop_name": "element",
             "table_name": "element"},
            
            {"prop_name": "mass",
             "table_name": "mass",
             "unit": lammps_unit['mass']},
            
            {"prop_name": "pos",
             "table_name": ["x", "y", "z"],
             "unit": lammps_unit['length']},
            
            {"prop_name": "spos",
             "table_name": ["xs", "ys", "zs"],
             "unit": 'scaled'},
            
            {"prop_name": "upos",
             "table_name": ["xu", "yu", "zu"],
             "unit": lammps_unit['length']},
            
            {"prop_name": "supos",
             "table_name": ["xsu", "ysu", "zsu"],
             "unit": 'scaled'},
            
            {"prop_name": "boximage",
             "table_name": ["ix", "iy", "iz"],
             "unit": 'scaled'},
            
            {"prop_name": "velocity",
             "table_name": ["vx", "vy", "vz"],
             "unit": lammps_unit['velocity']},
            
            {"prop_name": "force",
             "table_name": ["fx", "fy", "fz"],
             "unit": lammps_unit['force']},
            
            {"prop_name": "charge",
             "table_name": "q",
             "unit": lammps_unit['charge']},
            
            {"prop_name": "mu",
             "table_name": ["mux", "muy", "muz"],
             "unit": lammps_unit['dipole']},
            
            {"prop_name": "mu_mag",
             "table_name": "mu",
             "unit": lammps_unit['dipole']},
            
            {"prop_name": "radius",
             "table_name": "radius",
             "unit": lammps_unit['length']},
            
            {"prop_name": "diameter",
             "table_name": "diameter",
             "unit": lammps_unit['length']},
            
            {"prop_name": "ang_velocity",
             "table_name": ["omegax", "omegay", "omegaz"],
             "unit": lammps_unit['ang-vel']},
            
            {"prop_name": "ang_momentum",
             "table_name": ["angmomx", "angmomy", "angmomz"],
             "unit": lammps_unit['ang-mom']},
            
            {"prop_name": "torque",
             "table_name": ["tqx", "tqy", "tqz"],
             "unit": lammps_unit['force'] + '*' + lammps_unit['length']}]


# coding: utf-8

# http://www.numpy.org/
import numpy as np
import numpy.typing as npt

class Shape():
    """
    Template class for defining geometric regions in space.
    """

    def __init__(self):
        """
        Defines a shape
        """
        raise NotImplementedError('Base shape class cannot be directly used')
        
    def inside(self,
               pos: npt.ArrayLike,
               inclusive: bool = True) -> np.ndarray:
        """
        Indicates if position(s) are inside the shape.
        
        Parameters
        ----------
        pos : array-like object
            Nx3 array of coordinates. 
        inclusive : bool, optional
            Indicates if points on the shape's boundaries are to be included.
            Default value is True.
        
        Returns
        -------
        numpy.NDArray
            N array of bool values: True if inside shape
        """
        raise NotImplementedError('Base shape class cannot be directly used')
        
    def outside(self,
                pos: npt.ArrayLike,
                inclusive: bool = False) -> np.ndarray:
        """
        Indicates if position(s) are inside the shape.
        
        Parameters
        ----------
        pos : array-like object
            Nx3 array of coordinates. 
        inclusive : bool, optional
            Indicates if points on the shape's boundaries are to be included.
            Default value is False.
        
        Returns
        -------
        numpy.NDArray
            N array of bool values: True if outside shape
        """
        # Call inverse of inside
        return ~self.inside(pos, inclusive=not inclusive)
        


# coding: utf-8

# Standard Python imports
import io
from typing import Optional, Union, Tuple

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

# https://github.com/usnistgov/yabadaba
from yabadaba.record import Record
from yabadaba import load_query

# https://pandas.pydata.org/
import pandas as pd

# atomman imports
from ... import System

class CrystalPrototype(Record):
    """
    Class for representing crystal_prototype records that describe common
    crystal prototypes.
    """
    def __init__(self,
                 model: Union[str, io.IOBase, DM, None] = None,
                 name: Optional[str] = None):
        """
        Initializes a Record object for a given style.
        
        Parameters
        ----------
        model : str, file-like object, DataModelDict
            The contents of the record.
        name : str, optional
            The unique name to assign to the record.  If model is a file
            path, then the default record name is the file name without
            extension.
        """
        if model is not None:
            super().__init__(model=model, name=name)
        elif name is not None:
            self.name = name

    @property
    def style(self) -> str:
        """str: The record style"""
        return 'crystal_prototype'

    @property
    def modelroot(self) -> str:
        """str: The root element of the content"""
        return 'crystal-prototype'

    @property
    def xsd_filename(self) -> Tuple[str, str]:
        """tuple: The module path and file name of the record's xsd schema"""
        return ('atomman.library.xsd', f'{self.style}.xsd')

    def load_model(self,
                   model: Union[str, io.IOBase, DM],
                   name: Optional[str] = None):
        """
        Loads record contents from a given model.

        Parameters
        ----------
        model : str or DataModelDict
            The model contents of the record to load.
        name : str, optional
            The name to assign to the record.  Often inferred from other
            attributes if not given.
        """
        super().load_model(model, name=name)        
        proto = self.model[self.modelroot]
        
        self.__key = proto['key']
        self.__id = proto['id']
        self.__commonname = proto['name']
        self.__prototype = proto['prototype']
        self.__pearson = proto['Pearson-symbol']
        self.__strukturbericht = proto['Strukturbericht']
        
        self.__sg_number = proto['space-group']['number']
        self.__sg_hm = proto['space-group']['Hermann-Maguin']
        self.__sg_schoenflies = proto['space-group']['Schoenflies']
        
        self.__crystalfamily = proto['system-info']['cell']['crystal-family']
        self.__natypes = proto['system-info']['cell']['natypes']
        self.__ucell = None

        # Set name as id if no name given
        try:
            self.name
        except:
            self.name = self.id

    @property
    def id(self) -> str:
        """str : A unique id assigned to the record"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__id

    @property
    def key(self) -> str:
        """str : A UUID4 key assigned to the record"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__key

    @property
    def commonname(self) -> str:
        """str : A common name associated with the prototype"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__commonname
    
    @property
    def prototype(self) -> str:
        """str : A prototype composition associated with the prototype"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__prototype
    
    @property
    def pearson(self) -> str:
        """str : The prototype's Pearson symbol"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__pearson

    @property
    def strukturbericht(self) -> str:
        """str : The prototype's Strukturbericht symbol"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__strukturbericht

    @property
    def sg_number(self) -> int:
        """int : The prototype's space group number"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__sg_number

    @property
    def sg_hm(self) -> str:
        """str : The prototype's space group international symbol"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__sg_hm

    @property
    def sg_schoenflies(self) -> str:
        """str : The prototype's space group Schoenflies symbol"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__sg_schoenflies

    @property
    def crystalfamily(self) -> str:
        """str : The prototype's system family"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__crystalfamily

    @property
    def natypes(self) -> int:
        """int : Number of atom types"""
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.__natypes

    @property
    def ucell(self) -> System:
        """atomman.System : The unit cell for the prototype""" 
        if self.model is None:
            raise AttributeError('No model information loaded')
        if self.__ucell is None:
            self.__ucell = System(model=self.model)
        return self.__ucell

    def build_model(self) -> DM:
        if self.model is None:
            raise AttributeError('No model information loaded')
        return self.model

    def metadata(self) -> dict:
        """
        Generates a dict of simple metadata values associated with the record.
        Useful for quickly comparing records and for building pandas.DataFrames
        for multiple records of the same style.
        """
        params = {}
        params['name'] = self.name
        params['key'] = self.key
        params['id'] = self.id
        params['commonname'] = self.commonname
        params['prototype'] = self.prototype
        params['pearson'] = self.pearson
        params['strukturbericht'] = self.strukturbericht
        params['sg_number'] = self.sg_number
        params['sg_hm'] = self.sg_hm
        params['sg_schoenflies'] = self.sg_schoenflies
        params['crystalfamily'] = self.crystalfamily
        params['natypes'] = self.natypes
        
        return params

    @property
    def queries(self) -> dict:
        """dict: Query objects and their associated parameter names."""
        return {
            'key': load_query(
                style='str_match',
                name='key', 
                path=f'{self.modelroot}.key'),
            'id': load_query(
                style='str_match',
                name='id',
                path=f'{self.modelroot}.id'),
            'commonname': load_query(
                style='str_match',
                name='commonname',
                path=f'{self.modelroot}.name'),
            'prototype': load_query(
                style='str_match',
                name='prototype',
                path=f'{self.modelroot}.prototype'),
            'pearson': load_query(
                style='str_match',
                name='pearson',
                path=f'{self.modelroot}.Pearson-symbol'),
            'strukturbericht': load_query(
                style='str_match',
                name='strukturbericht',
                path=f'{self.modelroot}.Strukturbericht'),
            'sg_number': load_query(
                style='int_match',
                name='sg_number',
                path=f'{self.modelroot}.space-group.number'),
            'sg_hm': load_query(
                style='str_match',
                name='sg_hm',
                path=f'{self.modelroot}.space-group.Hermann-Maguin'),
            'sg_schoenflies': load_query(
                style='str_match',
                name='sg_schoenflies',
                path=f'{self.modelroot}.space-group.Schoenflies'),
            'crystalfamily': load_query(
                style='str_match',
                name='crystalfamily',
                path=f'{self.modelroot}.system-info.cell.crystal-family'),
            'natypes': load_query(
                style='int_match',
                name='natypes',
                path=f'{self.modelroot}.system-info.cell.natypes'),
        }

    def pandasfilter(self,
                     dataframe: pd.DataFrame,
                     name: Union[str, list, None] = None,
                     id: Union[str, list, None] = None,
                     key: Union[str, list, None] = None,
                     commonname: Union[str, list, None] = None,
                     prototype: Union[str, list, None] = None,
                     pearson: Union[str, list, None] = None,
                     strukturbericht: Union[str, list, None] = None,
                     sg_number: Union[int, list, None] = None,
                     sg_hm: Union[str, list, None] = None,
                     sg_schoenflies: Union[str, list, None] = None,
                     crystalfamily: Union[str, list, None] = None,
                     natypes: Union[int, list, None] = None) -> pd.Series:
        """
        Filters a pandas.DataFrame based on kwargs values for the record style.
        
        Parameters
        ----------
        dataframe : pandas.DataFrame
            A table of metadata for multiple records of the record style.
        name : str or list
            The record name(s) to parse by.
        id : str or list
            The record id(s) to parse by.
        key : str or list
            The record key(s) to parse by.
        commonname : str or list
            Prototype common name(s) to parse by.
        prototype : str or list
            Prototype composition(s) to parse by.
        pearson : str or list
            Pearson symbol(s) to parse by.
        strukturbericht : str or list
            Prototype Strukturbericht symbol(s) to parse by.
        sg_number : int or list
            Space group number(s) to parse by.
        sg_hm : str or list
            Space group international symbol(s) to parse by.
        sg_schoenflies : str or list
            Space group Schoenflies symbol(s) to parse by.
        crystalfamily : str or list
            Crystal structure families to parse by.
        natypes : int or list
            Number of atom types to parse by.

        Returns
        -------
        pandas.Series
            Boolean map of matching values
        """
        matches = super().pandasfilter(dataframe, name=name, id=id, key=key,
                                       commonname=commonname, prototype=prototype,
                                       pearson=pearson, strukturbericht=strukturbericht,
                                       sg_number=sg_number, sg_hm=sg_hm,
                                       sg_schoenflies=sg_schoenflies,
                                       crystalfamily=crystalfamily, natypes=natypes)
        return matches

    def mongoquery(self, 
                   name: Union[str, list, None] = None,
                   id: Union[str, list, None] = None,
                   key: Union[str, list, None] = None,
                   commonname: Union[str, list, None] = None,
                   prototype: Union[str, list, None] = None,
                   pearson: Union[str, list, None] = None,
                   strukturbericht: Union[str, list, None] = None,
                   sg_number: Union[int, list, None] = None,
                   sg_hm: Union[str, list, None] = None,
                   sg_schoenflies: Union[str, list, None] = None,
                   crystalfamily: Union[str, list, None] = None,
                   natypes: Union[int, list, None] = None) -> dict:
        """
        Builds a Mongo-style query based on kwargs values for the record style.
        
        Parameters
        ----------
        name : str or list
            The record name(s) to parse by.
        id : str or list
            The record id(s) to parse by.
        key : str or list
            The record key(s) to parse by.
        commonname : str or list
            Prototype common name(s) to parse by.
        prototype : str or list
            Prototype composition(s) to parse by.
        pearson : str or list
            Pearson symbol(s) to parse by.
        strukturbericht : str or list
            Prototype Strukturbericht symbol(s) to parse by.
        sg_number : int or list
            Space group number(s) to parse by.
        sg_hm : str or list
            Space group international symbol(s) to parse by.
        sg_schoenflies : str or list
            Space group Schoenflies symbol(s) to parse by.
        crystalfamily : str or list
            Crystal structure families to parse by.
        natypes : int or list
            Number of atom types to parse by.
        
        Returns
        -------
        dict
            The Mongo-style query
        """     
        mquery = super().mongoquery(name=name, id=id, key=key,
                                    commonname=commonname, prototype=prototype,
                                    pearson=pearson, strukturbericht=strukturbericht,
                                    sg_number=sg_number, sg_hm=sg_hm,
                                    sg_schoenflies=sg_schoenflies,
                                    crystalfamily=crystalfamily, natypes=natypes)
        return mquery

    def cdcsquery(self, 
                  id: Union[str, list, None] = None,
                  key: Union[str, list, None] = None,
                  commonname: Union[str, list, None] = None,
                  prototype: Union[str, list, None] = None,
                  pearson: Union[str, list, None] = None,
                  strukturbericht: Union[str, list, None] = None,
                  sg_number: Union[int, list, None] = None,
                  sg_hm: Union[str, list, None] = None,
                  sg_schoenflies: Union[str, list, None] = None,
                  crystalfamily: Union[str, list, None] = None,
                  natypes: Union[int, list, None] = None) -> dict:
        """
        Builds a CDCS-style query based on kwargs values for the record style.
        
        Parameters
        ----------
        id : str or list
            The record id(s) to parse by.
        key : str or list
            The record key(s) to parse by.
        commonname : str or list
            Prototype common name(s) to parse by.
        prototype : str or list
            Prototype composition(s) to parse by.
        pearson : str or list
            Pearson symbol(s) to parse by.
        strukturbericht : str or list
            Prototype Strukturbericht symbol(s) to parse by.
        sg_number : int or list
            Space group number(s) to parse by.
        sg_hm : str or list
            Space group international symbol(s) to parse by.
        sg_schoenflies : str or list
            Space group Schoenflies symbol(s) to parse by.
        crystalfamily : str or list
            Crystal structure families to parse by.
        natypes : int or list
            Number of atom types to parse by.
        
        Returns
        -------
        dict
            The CDCS-style query
        """
        mquery = super().cdcsquery(id=id, key=key,
                                    commonname=commonname, prototype=prototype,
                                    pearson=pearson, strukturbericht=strukturbericht,
                                    sg_number=sg_number, sg_hm=sg_hm,
                                    sg_schoenflies=sg_schoenflies,
                                    crystalfamily=crystalfamily, natypes=natypes)
        return mquery